import logging

from django.contrib.postgres.fields import DateRangeField, DateTimeRangeField
from django.db import models
from django.db.models.fields.related import ManyToManyRel, ManyToOneRel, OneToOneRel
from django_filters.filterset import FilterSetMetaclass, remote_queryset, settings
from django_filters.rest_framework import FilterSet as DjangoFilterSet
from wbcore.filters import fields
from wbcore.filters.fields.multiple_lookups import MultipleLookupFilter
from wbcore.signals.filters import add_filters

from .utils import check_required_filters

logger = logging.getLogger(__name__)


def _is_number(field):
    return (
        issubclass(field.__class__, models.DecimalField)
        or issubclass(field.__class__, models.FloatField)
        or issubclass(field.__class__, models.IntegerField)
    )


class CustomFilterSetMetaClass(FilterSetMetaclass):
    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)
        if _meta := getattr(new_class, "Meta", None):
            for parent_field_name, child_fields in getattr(_meta, "flatten_fields", dict()).items():
                if remote_field := getattr(_meta.model._meta.get_field(parent_field_name), "remote_field", None):
                    for child_field_name, lookup_exps in child_fields.items():
                        child_field = remote_field.model._meta.get_field(child_field_name)
                        if _is_number(child_field):
                            nested_field_name = f"{parent_field_name}__{child_field_name}"
                            for lookup_exp in lookup_exps:
                                new_class.base_filters[f"{nested_field_name}__{lookup_exp}"] = fields.NumberFilter(
                                    field_name=nested_field_name,
                                    lookup_expr=lookup_exp,
                                    precision=2,
                                    label=child_field.verbose_name,
                                )

                        # TODO extend with other filter fields type
        return new_class

    @classmethod
    def get_declared_filters(cls, bases, attrs):
        filters = super().get_declared_filters(bases, attrs)

        multi_filters: list[tuple[str, MultipleLookupFilter]] = [
            (filter_name, attrs.pop(filter_name))
            for filter_name, obj in list(attrs.items())
            if isinstance(obj, MultipleLookupFilter)
        ]

        for field_name, filter in multi_filters:
            filters |= filter.get_filters(field_name)

        return filters


class FilterSet(DjangoFilterSet, metaclass=CustomFilterSetMetaClass):
    FILTER_DEFAULTS = {
        models.BooleanField: {"filter_class": fields.BooleanFilter},
        models.NullBooleanField: {"filter_class": fields.BooleanFilter},
        models.CharField: {"filter_class": fields.CharFilter},
        models.TextField: {"filter_class": fields.CharFilter},
        models.SlugField: {"filter_class": fields.CharFilter},
        models.EmailField: {"filter_class": fields.CharFilter},
        models.FilePathField: {"filter_class": fields.CharFilter},
        models.UUIDField: {"filter_class": fields.CharFilter},
        models.URLField: {"filter_class": fields.CharFilter},
        models.GenericIPAddressField: {"filter_class": fields.CharFilter},
        models.CommaSeparatedIntegerField: {"filter_class": fields.CharFilter},
        models.DateField: {"filter_class": fields.DateFilter},
        models.DateTimeField: {"filter_class": fields.DateTimeFilter},
        models.TimeField: {"filter_class": fields.TimeFilter},
        models.IntegerField: {"filter_class": fields.NumberFilter},
        models.FloatField: {"filter_class": fields.NumberFilter},
        models.DecimalField: {"filter_class": fields.NumberFilter},
        DateTimeRangeField: {"filter_class": fields.DateTimeRangeFilter},
        DateRangeField: {"filter_class": fields.DateRangeFilter},
        # models.DurationField: {"filter_class": DurationFilter},
        # models.SmallIntegerField: {"filter_class": NumberFilter},
        # models.AutoField: {"filter_class": NumberFilter},
        # models.PositiveIntegerField: {"filter_class": NumberFilter},
        # models.PositiveSmallIntegerField: {"filter_class": NumberFilter},
        # models.UUIDField: {"filter_class": UUIDFilter},
        # Forward relationships
        models.OneToOneField: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ForeignKey: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ManyToManyField: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
        # Reverse relationships
        OneToOneRel: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        ManyToOneRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
        ManyToManyRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
    }

    @classmethod
    def filter_class_for_remote_filter(cls):
        if hasattr(cls, "get_filter_class_for_remote_filter") and callable(cls.get_filter_class_for_remote_filter()):
            return cls.get_filter_class_for_remote_filter()
        return cls

    def __init__(self, data=None, queryset=None, *, request=None, prefix=None, view=None):
        super().__init__(data=data, queryset=queryset, request=request, prefix=prefix)
        remote_filters = add_filters.send(sender=self.filter_class_for_remote_filter(), request=request)
        self.view = view
        for _, res in remote_filters:
            if res:
                for remote_filter_key, remote_filter in res.items():
                    setattr(remote_filter, "column_field_name", remote_filter_key)
                    self.filters[remote_filter_key] = remote_filter

    @classmethod
    def filter_for_lookup(cls, field, lookup_type):
        if lookup_type == "exact" and getattr(field, "choices", None):
            filter_class, params = fields.ChoiceFilter, {"choices": field.choices}
        else:
            filter_class, params = super().filter_for_lookup(field, lookup_type)

        # Check if it is a decimal field:
        if hasattr(field, "decimal_places"):
            params["precision"] = field.decimal_places

        if hasattr(field, "verbose_name"):
            params["label"] = field.verbose_name

        return filter_class, params

    @classmethod
    def get_dependency_map(cls):
        if hasattr(cls, "Meta"):
            yield from getattr(cls.Meta, "dependency_map", dict()).items()

    @classmethod
    def get_filters(cls):
        filters = super().get_filters()
        remote_filters = add_filters.send(sender=cls.filter_class_for_remote_filter())
        for _, res in remote_filters:
            if res:
                for remote_filter_key, remote_filter in res.items():
                    setattr(remote_filter, "column_field_name", remote_filter_key)
                    filters[remote_filter_key] = remote_filter

        for field, help_text in getattr(cls, "help_texts", {}).items():
            filters[field].help_text = help_text

        # Get dependency map specified in Meta and append the argument accordingly
        for field, values in cls.get_dependency_map():
            for value in values:
                filters[field].depends_on.append({"field": value, "options": {}})
        return filters

    def extract_required_field_labels(self):
        return [label for label, filter in self.base_filters.items() if getattr(filter, "required", False)]

    def filter_queryset(self, queryset):
        queryset = super().filter_queryset(queryset)
        check_required_filters(self.request, self.view, self)
        return queryset

    def fake_filter(self, queryset, name, value):
        return queryset



from django.forms import ValidationError
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.example_app.models import Stadium
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer


class StadiumModelSerializer(serializers.ModelSerializer):
    _city = GeographyRepresentationSerializer(filter_params={"level": 3}, source="city")
    total_capacity = serializers.DecimalField(decimal_places=0, max_digits=10, read_only=True)
    guest_rating = serializers.EmojiRatingField()

    def validate_guest_rating(self, value):
        if value < 0 or value > 4:
            raise ValidationError(_("Guest rating must be between 0 and 4."))
        return value

    @serializers.register_only_instance_resource()
    def teams_inline(self, instance: Stadium, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "teams_playing": reverse(
                "example_app:team-stadium-list", args=[instance.pk], request=self.context["request"]
            )
        }

    @serializers.register_only_instance_resource()
    def matches_inline(self, instance: Stadium, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "recent_matches": reverse(
                "example_app:match-stadium-list", args=[instance.pk], request=self.context["request"]
            )
        }

    class Meta:
        model = Stadium
        fields = (
            "id",
            "name",
            "total_capacity",
            "seating_capacity",
            "standing_capacity",
            "guest_rating",
            "city",
            "_city",
        )


class StadiumRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:stadium-detail")

    class Meta:
        model = Stadium
        fields = ("id", "name", "_detail")



from datetime import timedelta
from unittest.mock import patch

import pytest
from django.utils import timezone
from psycopg.types.range import DateRange
from rest_framework.test import APIRequestFactory
from wbcore.contrib.example_app.factories import (
    EventFactory,
    MatchFactory,
    PlayerFactory,
    SportPersonFactory,
    TeamFactory,
)
from wbcore.contrib.example_app.viewsets import (
    PlayerStatisticsChartModelViewSet,
    SportPersonRepresentationViewSet,
)
from wbcore.filters.defaults import current_year_date_range
from wbcore.test.utils import get_or_create_superuser


@pytest.mark.django_db
class TestEvent:
    @patch("wbcore.contrib.example_app.models.Match.reschedule_task")
    def test_filter_by_period(self, mock_reschedule):
        event = EventFactory(match__date_time=timezone.now())
        before_event = DateRange((timezone.now() - timedelta(7)).date(), (timezone.now() - timedelta(1)).date())
        after_event = DateRange((timezone.now() + timedelta(1)).date(), (timezone.now() + timedelta(7)).date())
        mvs = PlayerStatisticsChartModelViewSet(kwargs={"player_id": event.person})
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert len(mvs.filterset_class(request=request).filter_by_period(qs, "", current_year_date_range())) == 1
        assert (
            mvs.filterset_class(request=request).filter_by_period(qs, "", current_year_date_range()).first() == event
        )
        assert not mvs.filterset_class(request=request).filter_by_period(qs, "", before_event).exists()
        assert not mvs.filterset_class(request=request).filter_by_period(qs, "", after_event).exists()


@pytest.mark.django_db
class TestPerson:
    @patch("wbcore.contrib.example_app.models.Match.reschedule_task")
    def test_filter_by_match(self, mock_reschedule):
        ref1 = SportPersonFactory()
        match1 = MatchFactory(referee=ref1, home__coach=None, away__coach=None)
        ref2 = SportPersonFactory()
        home_coach = SportPersonFactory()
        away_coach = SportPersonFactory()
        home_team = TeamFactory(coach=home_coach)
        away_team = TeamFactory(coach=away_coach)
        home_player = PlayerFactory(current_team=home_team)
        away_player = PlayerFactory(current_team=away_team)
        match2 = MatchFactory(
            home=home_team,
            away=away_team,
            referee=ref2,
        )
        SportPersonFactory()
        PlayerFactory()
        mvs = SportPersonRepresentationViewSet()
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        qs = mvs.get_queryset()
        assert len(mvs.filterset_class(request=request).filter_by_match(qs, "", match1)) == 1
        assert mvs.filterset_class(request=request).filter_by_match(qs, "", match1).first() == ref1
        assert set(
            mvs.filterset_class(request=request).filter_by_match(qs, "", match2).values_list("id", flat=True)
        ) == {ref2.pk, home_coach.pk, away_coach.pk, home_player.pk, away_player.pk}



from contextlib import suppress

from wbcore.filters import DjangoFilterBackend

from .base import WBCoreViewConfig


class FilterFieldsViewConfig(WBCoreViewConfig):
    metadata_key = "filter_fields"
    config_class_attribute = "filter_fields_config_class"

    def get_metadata(self):
        filter_fields = dict()

        with suppress(StopIteration):
            backend = next(
                filter(lambda b: issubclass(b, DjangoFilterBackend), getattr(self.view, "filter_backends", []))
            )
            if filterset_class := backend().get_filterset_class(self.view, self.view.queryset):  # type: ignore
                filterset = filterset_class()
                filters = filterset.get_filters()
                hidden_fields = []
                # TODO: This should be the responsibility of the filterset_class to return also the df_fields
                if filterset_class_meta := getattr(filterset_class, "Meta", None):
                    hidden_fields.extend(getattr(filterset_class_meta, "hidden_fields", []))
                    filters.update(getattr(filterset_class_meta, "df_fields", {}))
                for name, field in filters.items():
                    field.parent = filterset
                    if (
                        not field.hidden
                        and name not in hidden_fields
                        and (res := field.get_representation(self.request, name, self.view))
                    ):
                        representation, lookup_expr = res
                        if field.key in filter_fields:
                            filter_fields[field.key]["lookup_expr"].append(lookup_expr)
                        else:
                            filter_fields[field.key] = representation
                            filter_fields[field.key]["lookup_expr"] = [lookup_expr]
                            filter_fields[field.key]["label"] = field.label

        return filter_fields



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Layout,
    Page,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default


class DataDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="label", label=_("Label")),
            dp.Field(key="workflow", label=_("Workflow")),
            dp.Field(key="data_type", label=_("Data Type")),
            dp.Field(key="required", label=_("Required")),
            dp.Field(key="help_text", label=_("Help Text")),
            dp.Field(key="default", label=_("Default")),
        ]

        if "workflow_id" in self.view.kwargs:
            fields.pop(1)

        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["label", "workflow", "data_type"],
                                ["required", "help_text", "default"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                    },
                ),
            ]
        )



from contextlib import suppress

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.content_type.utils import get_ancestors_content_type
from wbcore.contrib.authentication.authentication import JWTCookieAuthentication
from wbcore.contrib.documents.filters import DocumentFilter
from wbcore.contrib.documents.models import Document, DocumentModelRelationship
from wbcore.contrib.documents.serializers import (
    DocumentModelSerializer,
    DocumentRepresentationSerializer,
    ReadOnlyDocumentModelSerializer,
)
from wbcore.contrib.documents.viewsets.buttons import DocumentButtonConfig
from wbcore.contrib.documents.viewsets.display import DocumentModelDisplay
from wbcore.contrib.documents.viewsets.endpoints import DocumentEndpointConfig
from wbcore.contrib.documents.viewsets.previews import DocumentPreviewConfig
from wbcore.contrib.documents.viewsets.titles import DocumentModelTitleConfig
from wbcore.contrib.guardian.viewsets.mixins import GuardianFilterMixin


class DocumentModelViewSet(GuardianFilterMixin, viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "documents/markdown/documentation/documents.md"
    search_fields = ("name",)
    ordering = ["-updated", "name"]
    ordering_fields = ("name", "document_type", "updated")
    filterset_class = DocumentFilter
    preview_config_class = DocumentPreviewConfig
    queryset = Document.objects.all()
    serializer_class = DocumentModelSerializer
    title_config_class = DocumentModelTitleConfig
    endpoint_config_class = DocumentEndpointConfig
    button_config_class = DocumentButtonConfig
    display_config_class = DocumentModelDisplay

    def get_queryset(self):
        if (content_type_id := self.kwargs.get("content_type")) and (object_id := self.kwargs.get("content_id")):
            content_types = list(get_ancestors_content_type(ContentType.objects.get_for_id(content_type_id)))
            return (
                super()
                .get_queryset()
                .filter(
                    relationships__content_type__in=content_types,
                    relationships__object_id=object_id,
                )
            )
        return super().get_queryset()

    def create(self, request, *args, **kwargs):
        response = super().create(request, *args, **kwargs)

        if (content_type_id := self.kwargs.get("content_type")) and (object_id := self.kwargs.get("content_id")):
            DocumentModelRelationship.objects.create(
                document_id=response.data["instance"]["id"],
                content_type_id=content_type_id,
                object_id=object_id,
            )

        return response

    def get_serializer_class(self):
        if "pk" in self.kwargs:
            if (document := self.get_object()) and document.system_created:
                return ReadOnlyDocumentModelSerializer
        return super().get_serializer_class()

    @action(detail=True, methods=["PATCH"])
    def sendmail(self, request, pk=None):
        document = get_object_or_404(Document, id=pk)
        if to_email := request.POST.get("to_email", None):
            document.send_email(to_email)
            return Response(
                {"__notification": {"title": _("Document is going to be sent with email")}},
                status=status.HTTP_200_OK,
            )
        return Response(status=status.HTTP_400_BAD_REQUEST)

    @action(detail=False, methods=["GET"], authentication_classes=[JWTCookieAuthentication])
    def urlredirect(self, request, pk=None):
        """
        List Action to redirect the document url when a unique relationship is found for the given content type pair values and optional query parameters
        """
        filter_kwargs = {"system_created": request.GET.get("system_created", "true") == "true"}
        if "system_key" in request.GET:
            filter_kwargs["system_key"] = request.GET["system_key"]

        with suppress(Document.DoesNotExist, ObjectDoesNotExist, ContentType.DoesNotExist):
            content_type = ContentType.objects.get_for_id(self.request.GET.get("content_type"))
            obj = content_type.get_object_for_this_type(id=self.request.GET.get("object_id"))
            document = Document.get_for_object(obj).get(**filter_kwargs)
            if self.request.user.has_perm(Document.view_perm_str, document):
                return HttpResponseRedirect(document.file.url)
        return HttpResponse("Document could not be found", status=status.HTTP_404_NOT_FOUND)


class DocumentRepresentationViewSet(GuardianFilterMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcore:documents:documentrepresentation"
    serializer_class = DocumentRepresentationSerializer
    queryset = Document.objects.all()
    search_fields = ("name",)



from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from wbcore.configs.registry import ConfigRegistry


class ConfigAPIView(APIView):
    permission_classes = []

    def get(self, request: Request) -> Response:
        return Response(ConfigRegistry(request=request).get_config_dict())



import wbcore.serializers as wb_serializers
from django.utils.translation import gettext_lazy as _
from wbcore.utils.date import current_quarter_date_end, current_quarter_date_start


class StartEndDateSerializer(wb_serializers.Serializer):
    start = wb_serializers.DateField(label=_("Start"), default=current_quarter_date_start)
    end = wb_serializers.DateField(label=_("End"), default=current_quarter_date_end)



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class CalendarItemDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.Calendar(
            title="title",
            period="period",
            endpoint="endpoint",
            all_day="all_day",
            color="color",
            icon="icon",
            entities="entity_list",
        )



from .person import SportPersonTitleConfig, PlayerTitleConfig, PlayerTeamTitleConfig
from .sport import SportTitleConfig
from .league import LeagueTitleConfig, LeagueSportTitleConfig
from .stadium import StadiumTitleConfig
from .team import TeamTitleConfig, TeamStadiumTitleConfig
from .role import RoleTitleConfig
from .match import MatchTitleConfig, MatchStadiumTitleConfig, MatchLeagueTitleConfig
from .event import EventTitleConfig, EventTypeSportTitleConfig, EventTypeTitleConfig, PlayerStatisticsChartTitleConfig
from .teamresult import TeamResultsTitleConfig, TeamResultsLeagueTitleConfig



# Generated by Django 4.2.9 on 2024-02-20 10:08

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0006_employeremployeerelationship_position_name"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="bankingcontact",
            options={
                "permissions": (("administrate_bankingcontact", "Can Administrate Banking Contact"),),
                "verbose_name": "Banking Contact",
                "verbose_name_plural": "Banking Contacts",
            },
        ),
    ]



from contextlib import suppress
from importlib import import_module
from inspect import getmembers
from types import ModuleType
from typing import Any, Iterable

from django.conf import settings
from rest_framework.request import Request


class ConfigRegistry:
    """
    A class that is able to discover all functions that are wrapped in the `@register_config` decorator. This registry
    should never be used from the outside and its sole purpose is to gather all functions, so they are returned
    in the APIView of this module.
    """

    def __init__(self, request: Request, *args, **kwargs):
        self.request = request
        super().__init__(*args, **kwargs)

    def discover_configs(self) -> Iterable[ModuleType]:
        for app in settings.INSTALLED_APPS:
            with suppress(ModuleNotFoundError):
                module = import_module(f"{app}.configs")
                yield module

    def get_configs(self) -> Iterable[tuple[str, Any]]:
        for module in self.discover_configs():
            for member in getmembers(module, lambda member: hasattr(member, "_is_config")):
                if res := member[1](request=self.request):
                    yield res

    def get_config_dict(self) -> dict[str, Any]:
        return dict(self.get_configs())



from rest_framework.request import Request
from rest_framework.reverse import reverse
from wbcore.configs.decorators import register_config


@register_config
def gleap_config(request: Request) -> tuple[str, dict[str, str]]:
    return "gleap", {
        "user_identity_endpoint": reverse("gleap:user_identity", request=request),
        "api_token": reverse("gleap:api_token", request=request),
    }



from rest_framework import serializers

from .list import ListFieldMixin
from .mixins import WBCoreSerializerFieldMixin
from .types import WBCoreType


class ChoiceField(WBCoreSerializerFieldMixin, serializers.ChoiceField):
    field_type = WBCoreType.SELECT.value

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["choices"] = [{"value": k, "label": v} for k, v in self.choices.items()]
        return key, representation


class MultipleChoiceField(ListFieldMixin, WBCoreSerializerFieldMixin, serializers.MultipleChoiceField):
    field_type = WBCoreType.SELECT.value

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["multiple"] = True
        representation["choices"] = [{"value": k, "label": v} for k, v in self.choices.items()]
        return key, representation

    def to_internal_value(self, data):
        data = super().to_internal_value(data)
        if isinstance(data, set):
            data = list(data)
        return data

    def to_representation(self, data):
        data = super().to_representation(data)
        if isinstance(data, set):
            data = list(data)
        return data



from contextlib import suppress
from typing import Type

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from wbcore.metadata.configs.buttons import ButtonViewConfig
from wbcore.metadata.configs.display import (
    DisplayIdentifierViewConfig,
    DisplayViewConfig,
)
from wbcore.metadata.configs.documentations import DocumentationViewConfig
from wbcore.metadata.configs.endpoints import EndpointViewConfig
from wbcore.metadata.configs.fields import FieldsViewConfig
from wbcore.metadata.configs.filter_fields import FilterFieldsViewConfig
from wbcore.metadata.configs.identifiers import (
    DependantIdentifierViewConfig,
    IdentifierViewConfig,
)
from wbcore.metadata.configs.ordering_fields import OrderingFieldsViewConfig
from wbcore.metadata.configs.paginations import PaginationViewConfig
from wbcore.metadata.configs.preview import PreviewViewConfig
from wbcore.metadata.configs.primary_keys import PrimaryKeyViewConfig
from wbcore.metadata.configs.search_fields import SearchFieldsViewConfig
from wbcore.metadata.configs.titles import TitleViewConfig
from wbcore.metadata.configs.window_types import WindowTypeViewConfig

from .metadata import WBCoreMetadata


class WBCoreMetadataConfigViewMixin(
    FieldsViewConfig.as_view_mixin(),
    DocumentationViewConfig.as_view_mixin(),
    FilterFieldsViewConfig.as_view_mixin(),
    IdentifierViewConfig.as_view_mixin(),
    DependantIdentifierViewConfig.as_view_mixin(),
    PrimaryKeyViewConfig.as_view_mixin(),
    PaginationViewConfig.as_view_mixin(),
    WindowTypeViewConfig.as_view_mixin(),
    SearchFieldsViewConfig.as_view_mixin(),
    TitleViewConfig.as_view_mixin(),
    OrderingFieldsViewConfig.as_view_mixin(),
    EndpointViewConfig.as_view_mixin(),
    ButtonViewConfig.as_view_mixin(),
    DisplayViewConfig.as_view_mixin(),
    PreviewViewConfig.as_view_mixin(),
    DisplayIdentifierViewConfig.as_view_mixin(),
):
    metadata_class = WBCoreMetadata
    WIDGET_TYPE = None
    config_classes = [
        FieldsViewConfig,
        DocumentationViewConfig,
        FilterFieldsViewConfig,
        IdentifierViewConfig,
        DependantIdentifierViewConfig,
        PrimaryKeyViewConfig,
        PaginationViewConfig,
        WindowTypeViewConfig,
        SearchFieldsViewConfig,
        TitleViewConfig,
        OrderingFieldsViewConfig,
        EndpointViewConfig,
        ButtonViewConfig,
        DisplayViewConfig,
        PreviewViewConfig,
        DisplayIdentifierViewConfig,
    ]

    @property
    def new_mode(self):
        return self.request.GET.get("new_mode", "false") == "true" or self.action == "create"

    @property
    def inline(self):
        return self.request.GET.get("inline", "false").lower() == "true"

    def get_ordering_fields(self) -> list[str]:
        return getattr(self, "ordering_fields", [])

    def get_object(self) -> Model | None:
        # we cache the object in this method
        if hasattr(self, "_object"):
            return self._object
        else:
            with suppress(AttributeError):
                self._object = super().get_object()
                return self._object

    @classmethod
    def get_model(cls) -> Type[Model] | None:
        try:
            if hasattr(cls, "queryset") and cls.queryset is not None:
                return cls.queryset.model
            elif hasattr(cls, "serializer_class"):
                return cls.serializer_class.Meta.model
            else:
                return None
        except AttributeError:
            return None

    @classmethod
    def get_content_type(cls) -> ContentType | None:
        if model := cls.get_model():
            return ContentType.objects.get_for_model(model)
        return None



from collections import namedtuple

import pandas as pd
import pytest
from rest_framework.test import APIRequestFactory
from wbcore.pandas.filters import PandasOrderingFilter
from wbcore.viewsets import ViewSet


class TestPandasOrderingFilter:
    @pytest.fixture
    def ordering_filter(self):
        return PandasOrderingFilter()

    @pytest.fixture
    def view(self, base_ordering_fields, ordering_fields):
        class BaseView(ViewSet):
            def get_ordering_fields(self):
                return base_ordering_fields

            ordering = ordering_fields

        return BaseView()

    @pytest.fixture
    def df(self, ordering_fields):
        return pd.DataFrame(
            [[1] * len(ordering_fields), [2] * len(ordering_fields)],
            columns=map(lambda x: x.replace("-", ""), ordering_fields),
        )

    @pytest.fixture
    def queryset(self):
        # dummy iterable to assert that the filter queryset method doesn't touch the queryset
        return [namedtuple("Point", "x y")]

    @pytest.fixture
    def filter_request(self, ordering_fields):
        factory = APIRequestFactory()
        request = factory.get(f'/?ordering={",".join(ordering_fields)}')
        request.query_params = request.GET
        return request

    @pytest.mark.parametrize(
        "base_ordering_fields, ordering_fields, expected_ordering_fields, expected_ascending",
        [
            (["a", "b"], ["a"], ["a"], [True]),
            (["a", "b"], ["c"], [], []),
            (["a", "b"], ["-a"], ["a"], [False]),
            (["a", "b"], ["-a", "b"], ["a", "b"], [False, True]),
        ],
    )
    def test_get_ordering(
        self,
        base_ordering_fields,
        ordering_fields,
        expected_ordering_fields,
        expected_ascending,
        ordering_filter,
        filter_request,
        df,
        view,
    ):
        res_ordering_fields, res_ascending = ordering_filter.get_ordering(filter_request, df, view)
        assert list(res_ordering_fields) == expected_ordering_fields
        assert list(res_ascending) == expected_ascending

    @pytest.mark.parametrize(
        "base_ordering_fields, ordering_fields, expected_ordering_fields, expected_ascending",
        [
            (["a", "b"], ["a"], ["a"], [True]),
            (["a", "b"], ["c"], [], []),
            (["a", "b"], ["-a"], ["a"], [False]),
            (["a", "b"], ["-a", "b"], ["a", "b"], [False, True]),
        ],
    )
    def test_get_ordering_without_query_params(
        self,
        base_ordering_fields,
        ordering_fields,
        expected_ordering_fields,
        ordering_filter,
        expected_ascending,
        df,
        rf,
        view,
    ):
        rf.query_params = {}
        res_ordering_fields, res_ascending = ordering_filter.get_ordering(rf, df, view)
        assert list(res_ordering_fields) == expected_ordering_fields
        assert list(res_ascending) == expected_ascending

    @pytest.mark.parametrize(
        "base_ordering_fields, ordering_fields",
        [
            (["a", "b"], ["a"]),
        ],
    )
    def test_filter_queryset(
        self, base_ordering_fields, ordering_fields, ordering_filter, filter_request, queryset, view
    ):
        assert ordering_filter.filter_queryset(filter_request, queryset, view) == queryset

    @pytest.mark.parametrize(
        "base_ordering_fields, ordering_fields",
        [
            (["a", "b"], ["-a"]),
        ],
    )
    def test_filter_dataframe(self, base_ordering_fields, ordering_fields, ordering_filter, filter_request, df, view):
        assert list(ordering_filter.filter_dataframe(filter_request, df, view)["a"].values) == [
            2,
            1,
        ]  # we just check that ordering is applied



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons.buttons import ActionButton

from ..models.workflow import Workflow


def add_workflow_next_buttons_to_instance(sender, serializer, instance, request, user, **kwargs):
    buttons = []
    for transition_list, process_step in Workflow.get_next_user_step_transitions_for_instance(instance, request.user):
        for transition in transition_list:
            buttons.append(
                ActionButton(
                    method=RequestType.PATCH,
                    identifiers=("workflow:step",),
                    endpoint=f"{reverse('wbcore:workflow:processstep-next', args=[process_step.pk], request=request)}?transition_id={transition.pk}",
                    label=transition.name,
                    icon=transition.icon,
                    description_fields=_("Are you sure you want to activate {}?").format(transition.name),
                    title=transition.name,
                    action_label=_("Activating {}").format(transition.name),
                )
            )
    return buttons



from wbcore import viewsets
from wbcore.contrib.documents.models import DocumentModelRelationship
from wbcore.contrib.documents.serializers import (
    DocumentModelRelationshipModelSerializer,
)
from wbcore.contrib.documents.viewsets.display import (
    DocumentModelRelationshipViewConfig,
)
from wbcore.contrib.documents.viewsets.endpoints.documents_model_relationships import (
    DocumentModelRelationshipEndpointConfig,
)


class DocumentModelRelationshipModelViewSet(viewsets.ModelViewSet):
    queryset = DocumentModelRelationship.objects.all()
    serializer_class = DocumentModelRelationshipModelSerializer
    display_config_class = DocumentModelRelationshipViewConfig
    endpoint_config_class = DocumentModelRelationshipEndpointConfig

    def get_queryset(self):
        queryset = super().get_queryset()
        if document_id := self.kwargs.get("document_id", None):
            queryset = queryset.filter(document_id=document_id)

        return queryset



from enum import Enum

from django.forms import ValidationError
from django.utils.translation import gettext as _
from wbcore import serializers
from wbcore.contrib.example_app.models import League, Team, TeamResults
from wbcore.contrib.example_app.serializers import (
    LeagueRepresentationSerializer,
    TeamRepresentationSerializer,
)


class ResultErrorMessages(Enum):
    result_exists = _("Results for this team and league already exist.")


class TeamResultsModelSerializer(serializers.ModelSerializer):
    _team = TeamRepresentationSerializer(source="team")
    _league = LeagueRepresentationSerializer(source="league")
    games_played = serializers.IntegerField(label=_("Games"), default=0)

    def validate(self, data: dict) -> dict:
        team: Team | None = data.get("team", None)
        league: League | None = data.get("league", None)
        if team and league:
            team_results = TeamResults.objects.filter(team=team, league=league)
            if obj := self.instance:
                team_results = team_results.exclude(id=obj.id)
            if team_results.exists():
                raise ValidationError({"team": ResultErrorMessages.result_exists.value})
        return super().validate(data)

    class Meta:
        model = TeamResults
        fields = "__all__"


class TeamResultsRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:teamresults-detail")

    class Meta:
        model = TeamResults
        fields = ("id", "team", "league", "_detail")



from wbcore.metadata.configs.display.configs import preset_config

from .configs import (
    markdown_config,
    menu_calendar_config,
    menu_config,
    release_note_config,
    share_config,
    beta_button,
    frontend_version,
)



import pytest
from django.apps import apps
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcore.contrib.workflow.factories import (
    ConditionFactory,
    DataFactory,
    DataValueFactory,
    DecisionStepFactory,
    EmailStepFactory,
    FinishStepFactory,
    JoinStepFactory,
    ProcessFactory,
    ProcessStepFactory,
    RandomChildStepFactory,
    ScriptStepFactory,
    SplitStepFactory,
    StartStepFactory,
    StepFactory,
    TransitionFactory,
    UserStepFactory,
    WorkflowFactory,
)
from wbcore.tests.conftest import *

register(WorkflowFactory)
register(EmailStepFactory)
register(TransitionFactory)
register(DataFactory)
register(ProcessStepFactory)
register(UserStepFactory)
register(ConditionFactory)
register(RandomChildStepFactory)
register(FinishStepFactory)
register(ProcessFactory)
register(DecisionStepFactory)
register(SplitStepFactory)
register(JoinStepFactory)
register(ScriptStepFactory)
register(StepFactory)
register(DataValueFactory)
register(StartStepFactory)


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))



from django.utils.translation import gettext as _
from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.workflow.models import Step, Transition
from wbcore.contrib.workflow.serializers.step import StepRepresentationSerializer


class TransitionModelSerializer(wb_serializers.ModelSerializer):
    _from_step = StepRepresentationSerializer(source="from_step", required=False)

    _to_step = StepRepresentationSerializer(source="to_step", required=False)

    @wb_serializers.register_resource()
    def conditions_inline(self, instance: Transition, request, user) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "conditions": reverse(
                "wbcore:workflow:condition-transition-list", args=[instance.pk], request=self.context["request"]
            )
        }

    def validate(self, data: dict) -> dict:
        from_step: Step | None = data.get("from_step", self.instance.from_step if self.instance else None)
        to_step: Step | None = data.get("to_step", self.instance.to_step if self.instance else None)
        name: str | None = data.get("name", self.instance.name if self.instance else None)

        if to_step:
            if to_step.step_type == Step.StepType.STARTSTEP:
                raise ValidationError({"to_step": _("Cannot set a start step at the end of a transition.")})

            if from_step:
                if not from_step.workflow == to_step.workflow:
                    raise ValidationError({"to_step": _("All steps need to belong to the same workflow.")})
                if from_step == to_step:
                    raise ValidationError(
                        {
                            "to_step": _(
                                "Cannot transition to the same step. Please set different preceding and succeeding steps."
                            )
                        }
                    )
        if from_step:
            if from_step.step_type == Step.StepType.FINISHSTEP:
                raise ValidationError({"from_step": _("Cannot set a finishing step at the start of a transition.")})

        transition = Transition.objects.filter(name=name, to_step=to_step)
        if self.instance:
            transition = transition.exclude(id=self.instance.id)
        if transition.exists():
            raise ValidationError(
                {"name": _("A transition with this name to the same succeeding step already exists.")}
            )
        return data

    class Meta:
        model = Transition
        fields = (
            "id",
            "name",
            "from_step",
            "_from_step",
            "to_step",
            "_to_step",
            "icon",
            "_additional_resources",
        )


class TransitionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:transition-detail")

    class Meta:
        model = Transition
        fields = (
            "id",
            "name",
            "_detail",
        )



from wbcore.configs.registry import ConfigRegistry


def test_notifications_config(config_registry: ConfigRegistry):
    notifications = config_registry.get_config_dict()["notifications"]
    assert notifications["endpoint"]
    assert notifications["token"]



# Generated by Django 4.2.7 on 2024-02-07 10:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("workflow", "0004_alter_userstep_assignee_method"),
    ]

    operations = [
        migrations.AlterField(
            model_name="userstep",
            name="assignee_method",
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name="Assignee Method"),
        ),
    ]



import re

from rest_framework.reverse import reverse
from wbcore.enums import WidgetType
from wbcore.utils.urls import get_parse_endpoint, get_urlencode_endpoint

from .base import WBCoreViewConfig


class EndpointViewConfig(WBCoreViewConfig):
    metadata_key = "endpoints"
    config_class_attribute = "endpoint_config_class"

    PK_FIELD = "id"
    DELETE_PK_FIELD = "id"
    UPDATE_PK_FIELD = "id"

    def get_endpoint(self, **kwargs):
        model = self.view.get_model()
        basename_method_name = (
            "get_representation_endpoint"
            if self.view.WIDGET_TYPE == WidgetType.SELECT.value
            else "get_endpoint_basename"
        )

        if model and (basename_method := getattr(model, basename_method_name, None)):
            basename = basename_method().replace(
                "-list", ""
            )  # we replace "-list" in case it's a representation endpoint which historically comes with the sufix "-list"
            is_list = kwargs.get("is_list", False)
            if self.instance and not is_list:
                return reverse(f"{basename}-detail", args=[self.view.kwargs.get("pk")], request=self.request)
            else:
                return reverse(f"{basename}-list", request=self.request)
        return None

    def get_instance_endpoint(self, **kwargs):
        return self.get_endpoint()

    def _get_instance_endpoint(self):
        if endpoint := self.get_instance_endpoint():
            from wbcore.viewsets import ViewSet

            content_type = self.view.get_content_type()
            if content_type:
                change_permission = f"{content_type.app_label}.view_{content_type.model}"
                permission = self.request.user.has_perm(change_permission)
            elif isinstance(self.view, ViewSet):
                permission = True
            if permission:
                if self.instance:
                    if not re.search(r"\/\d+\/$", endpoint) and (pk := self.view.kwargs.get("pk", None)):
                        endpoint += str(pk) + "/"
                else:
                    endpoint += "{{" + self.PK_FIELD + "}}/"
                return endpoint
        return None

    def get_list_endpoint(self, **kwargs):
        return self.get_endpoint(is_list=True)

    def _get_list_endpoint(self):
        return self.get_list_endpoint()

    def get_delete_endpoint(self, **kwargs):
        return self.get_endpoint()

    def _get_delete_endpoint(self, **kwargs):
        read_only = getattr(self.view, "READ_ONLY", False)
        content_type = self.view.get_content_type()

        if content_type is None:
            return None

        delete_permission = f"{content_type.app_label}.delete_{content_type.model}"

        if read_only or not self.request.user.has_perm(delete_permission):
            return None
        if endpoint := self.get_delete_endpoint():
            if self.instance:
                if not re.search(r"\/\d+\/$", endpoint) and (pk := self.view.kwargs.get("pk", None)):
                    return f"{endpoint}{pk}/"
                return endpoint
            else:
                return endpoint + "{{" + self.DELETE_PK_FIELD + "}}/"
        return None

    def get_create_endpoint(self, **kwargs):
        return self.get_endpoint(is_list=True)

    def _get_create_endpoint(self):
        from wbcore.viewsets import ViewSet

        read_only = getattr(self.view, "READ_ONLY", False)
        if read_only:
            return None

        if endpoint := self.get_create_endpoint():
            content_type = self.view.get_content_type()
            if content_type:
                create_permission = f"{content_type.app_label}.add_{content_type.model}"
                permission = self.request.user.has_perm(create_permission)
            elif isinstance(self.view, ViewSet):
                permission = True
            if permission:
                endpoint, query_params = get_parse_endpoint(endpoint)
                query_params["new_mode"] = "true"
                return get_urlencode_endpoint(endpoint, query_params)

        return None

    def get_update_endpoint(self, **kwargs):
        return self.get_instance_endpoint()

    def _get_update_endpoint(self):
        if (endpoint := self.get_update_endpoint()) and not getattr(self.view, "READ_ONLY", False):
            from wbcore.viewsets import ViewSet

            content_type = self.view.get_content_type()
            if content_type:
                change_permission = f"{content_type.app_label}.change_{content_type.model}"
                permission = self.request.user.has_perm(change_permission)
            elif isinstance(self.view, ViewSet):
                permission = True
            if permission:
                if self.instance:
                    if not re.search(r"\/\d+\/$", endpoint) and (pk := self.view.kwargs.get("pk", None)):
                        endpoint += str(pk) + "/"
                else:
                    endpoint += "{{" + self.UPDATE_PK_FIELD + "}}/"
                return endpoint
        return None

    def get_reorder_endpoint(self, **kwargs):
        return self._get_update_endpoint()

    def _get_reorder_endpoint(self):
        if (endpoint := self.get_reorder_endpoint()) and hasattr(self.view, "reorder"):
            return endpoint + "reorder/"

    def get_pre_change_endpoint(self, pk):
        return None

    def _get_pre_change_endpoint(self):
        if (pk := self.view.kwargs.get("pk", None)) and (endpoint := self.get_pre_change_endpoint(pk)):
            return endpoint
        return None

    def get_pre_create_endpoint(self, pk):
        return None

    def _get_pre_create_endpoint(self):
        if (pk := self.view.kwargs.get("pk", None)) and (endpoint := self.get_pre_create_endpoint(pk)):
            return endpoint
        return None

    def get_pre_delete_endpoint(self, pk):
        return None

    def _get_pre_delete_endpoint(self):
        if (pk := self.view.kwargs.get("pk", None)) and (endpoint := self.get_pre_delete_endpoint(pk)):
            return endpoint
        return None

    def get_metadata(self) -> dict:
        return {
            "instance": self._get_instance_endpoint(),
            "list": self._get_list_endpoint(),
            "delete": self._get_delete_endpoint(),
            "create": self._get_create_endpoint(),
            "update": self._get_update_endpoint(),
            "reorder": self._get_reorder_endpoint(),
            "pre_change": self._get_pre_change_endpoint(),
            "pre_create": self._get_pre_create_endpoint(),
            "pre_delete": self._get_pre_delete_endpoint(),
        }



from django.dispatch import receiver
from wbcore.test.signals import get_custom_factory

from ..factories import CalendarItemFactory
from ..viewsets import (
    CalendarItemViewSet,
    InfiniteCalendarItemViewSet,
    OwnCalendarItemViewSet,
)


@receiver(get_custom_factory, sender=CalendarItemViewSet)
@receiver(get_custom_factory, sender=InfiniteCalendarItemViewSet)
@receiver(get_custom_factory, sender=OwnCalendarItemViewSet)
def receive_factory_employee(sender, *args, **kwargs):
    return CalendarItemFactory



from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.fields.reverse_related import ManyToManyRel
from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from slugify import slugify
from wbcore.contrib.color.fields import ColorField
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin


class ManagedMixin(models.Model):
    managed = models.BooleanField(default=False)

    class Meta:
        abstract = True


class Tag(ManagedMixin, ComplexToStringMixin, WBModel):
    title = models.CharField(max_length=255)

    slug = models.CharField(max_length=255, null=True, blank=True)
    color = ColorField(default="#D3D3D3")

    groups = models.ManyToManyField(to="tags.TagGroup", blank=True)
    description = models.TextField(default="", blank=True)

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, blank=True, null=True, related_name="associated_tags"
    )

    class Meta:
        unique_together = [("slug", "content_type")]
        verbose_name = "Tag"
        verbose_name_plural = "Tags"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:tags:tag"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:tags:tagrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"

    def compute_str(self) -> str:
        # We need to check here if the instance already is in the database, otherwise groups cannot be linked yet
        if self.id and self.groups.exists():
            groups = ",".join(self.groups.all().values_list("title", flat=True))
            return f"{groups}::{self.title}"
        return self.title

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.title)
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    def get_tagged_queryset(self):
        qs = self.__class__.objects.none()
        for field in filter(lambda x: isinstance(x, ManyToManyRel), self.__class__._meta.get_fields()):
            _qs = getattr(self, field.related_name).all()
            qs = qs.union(_qs.values("tag_detail_endpoint", "tag_representation"))
        return qs


class TagGroup(ManagedMixin, WBModel):
    title = models.CharField(max_length=255, unique=True)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:tags:taggroup"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:tags:taggrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    def __str__(self) -> str:
        return f"{self.title}"

    class Meta:
        verbose_name = "Tag Group"
        verbose_name_plural = "Tag Groups"


@receiver(m2m_changed, sender=Tag.groups.through)
def tag_groups_m2m_changed(sender, instance, action, **kwargs):
    instance.save()



import pytest
from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin


@pytest.fixture
def mocked_permission_object_model_mixin(mocker):
    def __init__(self):
        ...

    PermissionObjectModelMixin.__init__ = __init__
    PermissionObjectModelMixin.objects = mocker.Mock()
    mocker.patch.object(PermissionObjectModelMixin, "view_perm_str", "view_perm_str")
    mocker.patch.object(PermissionObjectModelMixin, "change_perm_str", "change_perm_str")
    mocker.patch.object(PermissionObjectModelMixin, "delete_perm_str", "delete_perm_str")
    mocker.patch.object(PermissionObjectModelMixin, "select_perm_str", "select_perm_str")
    mocker.patch.object(PermissionObjectModelMixin, "add_perm_str", "add_perm_str")
    mocker.patch.object(PermissionObjectModelMixin, "admin_perm_str", "admin_perm_str")
    return PermissionObjectModelMixin


class TestPermissionObjectModelMixin:
    def test_save_run_assign_permissions(self, mocker, mocked_permission_object_model_mixin):
        mocker.patch("django.db.models.Model.save")
        on_commit = mocker.patch("wbcore.contrib.guardian.models.mixins.transaction.on_commit")
        ContentType = mocker.patch("wbcore.contrib.guardian.models.mixins.ContentType")
        ContentType.objects.get_for_model.return_value = mocker.Mock()

        mocked_permission_object_model_mixin().save()

        on_commit.assert_called_once()

    def test_reload_permissions(self, mocker, mocked_permission_object_model_mixin):
        reload_permissions = mocker.patch("wbcore.contrib.guardian.models.mixins.reload_permissions")
        mocked_permission_object_model_mixin().reload_permissions(prune_existing=True, force_pruning=True)
        reload_permissions.assert_called_once()

    def test_get_permissions_for_superuser(self, mocker, mocked_permission_object_model_mixin):
        user = mocker.Mock()
        user.is_superuser = True
        permissions = mocked_permission_object_model_mixin().get_permissions_for_user(user)
        assert permissions == {}

    def test_get_permissions_for_internal_admin_user(self, mocker, mocked_permission_object_model_mixin):
        user = mocker.Mock()
        user.is_internal = True
        user.is_superuser = False
        mocker.patch.object(user, "has_perm", side_effect=lambda x: x == "admin_perm_str")
        permissions = mocked_permission_object_model_mixin().get_permissions_for_user(user)
        assert permissions == {
            "view_perm_str": False,
            "change_perm_str": False,
            "delete_perm_str": False,
            "select_perm_str": False,
            "add_perm_str": False,
        }

    def test_get_permissions_for_external_user(self, mocker, mocked_permission_object_model_mixin):
        user = mocker.Mock()
        user.is_internal = False
        user.is_superuser = False
        permissions = mocked_permission_object_model_mixin().get_permissions_for_user(user)
        assert permissions == {}

    @pytest.mark.parametrize(
        "permission", ["view_perm_str", "change_perm_str", "delete_perm_str", "select_perm_str", "add_perm_str"]
    )
    def test_get_permission_for_internal_user_and_single_permission(
        self, mocker, mocked_permission_object_model_mixin, permission
    ):
        user = mocker.Mock()
        user.is_internal = True
        user.is_superuser = False

        mocker.patch.object(user, "has_perm", side_effect=lambda x: x == permission)

        permissions = mocked_permission_object_model_mixin().get_permissions_for_user(user)
        assert permissions == {permission: False}


@pytest.mark.django_db
def test_post_save_user(user_factory, mocker):
    on_commit = mocker.patch("wbcore.contrib.guardian.models.mixins.transaction.on_commit")
    user_factory.create()

    on_commit.assert_called_once()


@pytest.mark.django_db
def test_post_save_user_not_called(user_factory, mocker):
    user_factory.create()
    on_commit = mocker.patch("wbcore.contrib.guardian.models.mixins.transaction.on_commit")

    on_commit.assert_not_called()



from slugify import slugify
from wbcore.contrib.example_app.models import EventType


def get_event_types_for_league(league_id: int) -> list[dict[str, int | str]]:
    """Retrieves all event types involved inside a league

    Args:
        league_id (str): ID of the league in question

    Returns:
        list[dict[int,str, str]]: A list of dictionaries with the event type's id, name and slugified name
    """

    event_types = (
        EventType.objects.filter(sport__leagues=league_id, events__isnull=False).distinct().values("id", "name")
    )
    for event_type in event_types:
        event_type["slugified_name"] = slugify(event_type["name"], separator="_")
    return event_types



import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models import User
from wbcore.serializers import PrimaryKeyRelatedField, RepresentationSerializer


class TestRepresentationSerializer:
    @pytest.fixture()
    def test_representation_serializer(self):
        class TestFixtureRepresentationSerializer(RepresentationSerializer):
            filter_params = {"a": "a"}

            class Meta:
                model = User
                fields = ("id",)

        return TestFixtureRepresentationSerializer

    @pytest.fixture()
    def option_request(self):
        factory = APIRequestFactory()
        return factory.get("/")

    def test_filter_params(self, test_representation_serializer, option_request):
        serializer = test_representation_serializer(source="source")
        assert serializer._get_filter_params(option_request) == {"a": "a"}

        assert serializer.get_representation(option_request, "id")[1]["endpoint"]["url"].endswith("?a=a")
        # assert kwargs override class attribute
        serializer = test_representation_serializer(source="source", filter_params={"b": "b"})
        assert serializer._get_filter_params(option_request) == {"b": "b"}

        serializer = test_representation_serializer(source="source", ignore_filter=True)
        assert serializer._get_filter_params(option_request) == dict()

    def test_filter_params_with_method(self, option_request):
        class TestFixtureRepresentationSerializer(RepresentationSerializer):
            def get_filter_params(self, *args):
                return {"c": "c"}

            class Meta:
                model = User
                fields = ("id",)

        serializer = TestFixtureRepresentationSerializer(source="source", filter_params={"b": "b"})
        assert serializer._get_filter_params(option_request) == {
            "b": "b"
        }  # kwargs or class attributes take precedance

        serializer = TestFixtureRepresentationSerializer(source="source")
        assert serializer._get_filter_params(option_request) == {
            "c": "c"
        }  # kwargs or class attributes take precedance

    def test_queryset_unset_if_read_only_callable(self):
        u1 = UserFactory.build()
        UserFactory.build()  # noisy user
        queryset = [u1]
        # test that the queryset is properly unset if the readonly attribute is a callable and behave normally otherwise
        related_field = PrimaryKeyRelatedField(read_only=False, queryset=queryset)
        assert related_field.queryset == queryset
        with pytest.raises(AssertionError):
            PrimaryKeyRelatedField(read_only=True, queryset=queryset)

        related_field = PrimaryKeyRelatedField(read_only=lambda: True, queryset=queryset)
        assert related_field.queryset is None

    def test_related_field_binding_reinstate_queryset_on_readonly_callable(self, test_representation_serializer):
        u1 = UserFactory.build()
        UserFactory.build()  # noisy user
        queryset = [u1]
        parent = test_representation_serializer(context={"view": object()})
        related_field = PrimaryKeyRelatedField(read_only=lambda view: False, queryset=queryset)
        assert related_field.queryset is None
        related_field.bind("related_field", parent)
        assert related_field.queryset == queryset



from unittest.mock import patch

import pytest
from rest_framework.reverse import reverse
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.factories import BankingContactFactory, PersonFactory
from wbcore.contrib.workflow.models import ProcessStep


@pytest.mark.django_db
class TestUserStep:
    def test_get_assigned_group(self, user_step_factory):
        user_step = user_step_factory()
        assert user_step.get_assigned_group() == user_step.group

    def test_get_assigned_user(self, user_step_factory):
        user = UserFactory()
        user_step = user_step_factory(assignee=user, group=None)
        assert user_step.get_assigned_user() == user

    def test_execute_assignee_method(self, process_step_factory, user_step_factory):
        step = user_step_factory()
        manager = PersonFactory()
        manager_account = UserFactory(profile=manager)
        assignee = PersonFactory(relationship_managers=[manager])
        instance = BankingContactFactory(entry=assignee)
        process_step = process_step_factory(step=step, process__instance=instance)
        process_step.step.execute_assignee_method(
            process_step, "manager_of_instance_assignee", assignee_field="entry", assignee_type="entry"
        )
        assert process_step.assignee == manager_account

    @patch("wbcore.contrib.workflow.models.step.UserStep.execute_assignee_method")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_run(self, mock_notify, mock_execute, process_step_factory, transition_factory, user_step_factory):
        kwargs = {"test": "Test"}
        step = user_step_factory(notify_user=True, kwargs=kwargs)
        user = UserFactory(groups=[step.group])
        process_step = process_step_factory(state=ProcessStep.StepState.ACTIVE, step=step)
        transition_factory(from_step=step)
        step.run(process_step)
        assert mock_notify.call_args.args == ([user], process_step)
        assert mock_execute.call_args.args == (process_step, step.assignee_method)
        assert mock_execute.call_args.kwargs == kwargs

    @patch("wbcore.contrib.workflow.models.step.UserStep.execute_assignee_method")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_run_no_kwargs(
        self, mock_notify, mock_execute, process_step_factory, transition_factory, user_step_factory
    ):
        step = user_step_factory(notify_user=True)
        user = UserFactory(groups=[step.group])
        process_step = process_step_factory(state=ProcessStep.StepState.ACTIVE, step=step)
        transition_factory(from_step=step)
        step.run(process_step)
        assert mock_notify.call_args.args == ([user], process_step)
        assert mock_execute.call_args.args == (process_step, step.assignee_method)
        assert mock_execute.call_args.kwargs == {}

    @patch("wbcore.contrib.workflow.models.step.UserStep.execute_assignee_method")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_run_no_notification(
        self, mock_notify, mock_execute, process_step_factory, transition_factory, user_step_factory
    ):
        kwargs = {"test": "Test"}
        step = user_step_factory(notify_user=False, kwargs=kwargs)
        UserFactory(groups=[step.group])
        process_step = process_step_factory(state=ProcessStep.StepState.ACTIVE, step=step)
        transition_factory(from_step=step)
        step.run(process_step)
        assert not mock_notify.called
        assert mock_execute.call_args.args == (process_step, step.assignee_method)
        assert mock_execute.call_args.kwargs == kwargs

    @patch("wbcore.contrib.workflow.models.step.UserStep.execute_assignee_method")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_run_assignee_method_failed(
        self, mock_notify, mock_execute, process_step_factory, transition_factory, user_step_factory
    ):
        kwargs = {"test": "Test"}
        step = user_step_factory(notify_user=True, kwargs=kwargs)
        UserFactory(groups=[step.group])
        process_step = process_step_factory(state=ProcessStep.StepState.FAILED, step=step)
        transition_factory(from_step=step)
        step.run(process_step)
        assert not mock_notify.called
        assert mock_execute.call_args.args == (process_step, step.assignee_method)
        assert mock_execute.call_args.kwargs == kwargs

    @patch("wbcore.contrib.workflow.models.step.UserStep.set_failed")
    @patch("wbcore.contrib.workflow.models.step.UserStep.execute_assignee_method")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_run_no_assignee(
        self, mock_notify, mock_execute, mock_failed, process_step_factory, transition_factory, user_step_factory
    ):
        kwargs = {"test": "Test"}
        step = user_step_factory(notify_user=True, kwargs=kwargs)
        process_step = process_step_factory(state=ProcessStep.StepState.ACTIVE, step=step)
        transition_factory(from_step=step)
        step.run(process_step)
        assert not mock_notify.called
        assert mock_execute.call_args.args == (process_step, step.assignee_method)
        assert mock_execute.call_args.kwargs == kwargs
        assert mock_failed.call_args.args[0] == process_step

    @patch("wbcore.contrib.workflow.models.step.UserStep.set_failed")
    @patch("wbcore.contrib.workflow.models.step.UserStep.execute_assignee_method")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_run_no_transition(
        self, mock_notify, mock_execute, mock_failed, process_step_factory, transition_factory, user_step_factory
    ):
        kwargs = {"test": "Test"}
        step = user_step_factory(notify_user=True, kwargs=kwargs)
        UserFactory(groups=[step.group])
        process_step = process_step_factory(state=ProcessStep.StepState.ACTIVE, step=step)
        step.run(process_step)
        assert not mock_notify.called
        assert mock_execute.call_args.args == (process_step, step.assignee_method)
        assert mock_execute.call_args.kwargs == kwargs
        assert mock_failed.call_args.args[0] == process_step

    @patch("wbcore.contrib.workflow.models.step.process_can_finish.delay")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_set_failed_user_step_notify_user(
        self, mock_notify, mock_can_finish, process_step_factory, user_step_factory
    ):
        step = user_step_factory(notify_user=True)
        UserFactory(groups=[step.group])
        process_step = process_step_factory(step=step, error_message=None)
        error_message = "Error message"
        step.set_failed(process_step, error_message)
        assert process_step.state == ProcessStep.StepState.FAILED
        assert process_step.error_message == error_message
        assert mock_notify.call_args.args == (process_step.get_all_assignees(), process_step, error_message)
        assert mock_can_finish.call_args.args == (process_step.process.pk,)

    @patch("wbcore.contrib.workflow.models.step.process_can_finish.delay")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_set_failed_user_step_not_notify(
        self, mock_notify, mock_can_finish, process_step_factory, user_step_factory
    ):
        step = user_step_factory(notify_user=False)
        UserFactory(groups=[step.group])
        process_step = process_step_factory(step=step, error_message=None)
        error_message = "Error message"
        step.set_failed(process_step, error_message)
        assert process_step.state == ProcessStep.StepState.FAILED
        assert process_step.error_message == error_message
        assert not mock_notify.called
        assert mock_can_finish.call_args.args == (process_step.process.pk,)

    @patch("wbcore.contrib.workflow.models.step.process_can_finish.delay")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_set_failed_user_step_no_assignees(
        self, mock_notify, mock_can_finish, process_step_factory, user_step_factory
    ):
        step = user_step_factory(notify_user=True)
        process_step = process_step_factory(step=step, error_message=None)
        error_message = "Error message"
        step.set_failed(process_step, error_message)
        assert process_step.state == ProcessStep.StepState.FAILED
        assert process_step.error_message == error_message
        assert not mock_notify.called
        assert mock_can_finish.call_args.args == (process_step.process.pk,)

    @patch("wbcore.contrib.workflow.models.step.send_notification")
    def test_notify_single_assignee(self, mock_notification, process_step_factory, user_step_factory):
        step = user_step_factory()
        process_step = process_step_factory(step=step)
        assignee_list = [UserFactory()]
        step.notify_assignees(assignee_list, process_step)
        assert mock_notification.call_args.kwargs == {
            "code": "workflow.userstep.notify_next_step",
            "title": "Workflow Step Awaiting Your Decision",
            "body": "You were assigned to a workflow step. Please select the next step.",
            "user": assignee_list[0],
            "endpoint": reverse(f"{process_step.get_endpoint_basename()}-detail", args=[process_step.id]),
        }

    @patch("wbcore.contrib.workflow.models.step.send_notification")
    def test_notify_assignee_error(self, mock_notification, process_step_factory, user_step_factory):
        step = user_step_factory()
        process_step = process_step_factory(step=step)
        assignee_list = [UserFactory()]
        error_message = "Test error"
        step.notify_assignees(assignee_list, process_step, error_message)
        assert mock_notification.call_args.kwargs == {
            "code": "workflow.userstep.notify_failed_step",
            "title": "Assigned Workflow Step Failed",
            "body": f"A workflow step you were assigned to just failed with the error message '{error_message}'. Please take appropriate action.",
            "user": assignee_list[0],
            "endpoint": reverse(f"{process_step.get_endpoint_basename()}-detail", args=[process_step.id]),
        }

    @patch("wbcore.contrib.workflow.models.step.send_notification")
    def test_notify_assignee_list(self, mock_notification, process_step_factory, user_step_factory):
        step = user_step_factory()
        process_step = process_step_factory(step=step)
        assignee_list = UserFactory.create_batch(3)
        step.notify_assignees(assignee_list, process_step)
        assert mock_notification.call_count == 3

    @patch("wbcore.contrib.workflow.models.step.send_notification")
    def test_notify_assignee_inactive_user(self, mock_notification, process_step_factory, user_step_factory):
        step = user_step_factory()
        process_step = process_step_factory(step=step)
        assignee_list = UserFactory.create_batch(3)
        inactive_user = UserFactory(is_active=False)
        assignee_list.append(inactive_user)
        step.notify_assignees(assignee_list, process_step)
        assert mock_notification.call_count == 3
        for call in mock_notification.call_args_list:
            assert call[1]["user"] != inactive_user

    @patch("wbcore.contrib.workflow.models.step.send_notification")
    @patch("wbcore.contrib.workflow.models.step.Step.user_can_see_step")
    def test_notify_assignee_user_cant_see(
        self, mock_can_see, mock_notification, process_step_factory, user_step_factory
    ):
        step = user_step_factory()
        process_step = process_step_factory(step=step)
        assignee_list = [UserFactory()]
        mock_can_see.return_value = False
        step.notify_assignees(assignee_list, process_step)
        assert not mock_notification.called



from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.identifiers import IdentifierViewConfig


def add_clear_cache_button(sender, instance, request, view, pk=None, **kwargs):
    if request.user.is_superuser and not instance:
        cache_key = view._get_cache_key()
        identifier = IdentifierViewConfig(view, request, instance=instance).get_metadata()
        return bt.ActionButton(
            method=RequestType.PATCH,
            icon=WBIcon.DOWNLOAD.icon,
            identifiers=(identifier,),
            endpoint=reverse("wbcore:clear_cache", args=[cache_key], request=request),
            action_label=_("Clear Cache"),
            title=_("Clear Cache"),
            label=_("Clear Cache"),
            description_fields=_("Clear Cache"),
        )



from rest_framework.reverse import reverse
from reversion.models import Revision, Version
from wbcore import serializers
from wbcore import serializers as wb_serializers


class RevisionRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbcore:revision-detail")

    endpoint = "wbcore:revisionrepresentation-list"
    value_key = "id"
    label_key = "{{date_created}}"

    class Meta:
        model = Revision
        fields = ("id", "date_created", "_detail")


class VersionRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbcore:version-detail")
    user_repr = serializers.StringRelatedField(source="revision.user")
    revision_date_created = serializers.StringRelatedField(source="revision.date_created")

    endpoint = "wbcore:versionrepresentation-list"
    value_key = "id"
    label_key = "{{id}} - {{user_repr}} - {{revision_date_created}}"

    class Meta:
        model = Version
        fields = ("id", "user_repr", "revision_date_created", "_detail")


class RevisionModelSerializer(serializers.ModelSerializer):
    user_repr = serializers.StringRelatedField(read_only=True, source="user")
    user = serializers.IntegerField(read_only=True, source="user.id")
    # @wb_serializers.register_resource()
    # def versions(self, instance, request, user):
    #     return {"versions": reverse("wbcore:revision-version-list", args=[instance.id], request=request)}

    class Meta:
        model = Revision
        fields = read_only_fields = (
            "id",
            "date_created",
            "user_repr",
            "user",
            "comment",
            "_additional_resources",
        )


class VersionModelSerializer(serializers.ModelSerializer):
    date_created = serializers.DateTimeField(read_only=True, source="revision.date_created")
    _revision = RevisionRepresentationSerializer(source="revision")
    profile_repr = serializers.CharField(read_only=True)

    object_repr = serializers.CharField()

    @wb_serializers.register_resource()
    def resources(self, instance, request, user):
        res = {"compare_with": reverse("wbcore:version-comparewith", args=[instance.id], request=request)}
        if user.has_perm("reversion.change_version") or user.is_superuser:
            res["revert"] = reverse("wbcore:version-revert", args=[instance.id], request=request)
        return res

    class Meta:
        model = Version
        fields = read_only_fields = (
            "id",
            "profile_repr",
            "revision",
            "_revision",
            "content_type",
            "object_id",
            "object_repr",
            "serialized_data",
            "date_created",
            "_additional_resources",
            # "field_dict"
        )



import pytest
from django.db import IntegrityError
from wbcore.contrib.example_app.factories import (
    LeagueFactory,
    MatchFactory,
    PlayerFactory,
    RoleFactory,
    SportFactory,
    SportPersonFactory,
    StadiumFactory,
    TeamFactory,
    TeamResultsFactory,
)
from wbcore.contrib.example_app.models import League, Match, Player, SportPerson


@pytest.mark.django_db
class TestPerson:
    def test_str_full(self):
        person = SportPersonFactory(first_name="Test", last_name="Name")
        assert person.str_full() == "Test Name"

    def test_str_full_coach(self):
        person = SportPersonFactory(first_name="Test", last_name="Name")
        TeamFactory(coach=person, name="Test Team")
        assert person.str_full() == "Test Name (Coach of Test Team)"

    def test_str_full_commissioner(self):
        commissioner = SportPersonFactory(first_name="Test", last_name="Name")
        LeagueFactory(commissioner=commissioner, name="Test League")
        assert commissioner.str_full() == "Test Name (Commissioner of Test League)"

    def test_get_player_fail(self):
        person = SportPersonFactory()
        assert person.get_player() is None

    def test_get_player_sucess(self):
        player = PlayerFactory()
        person = SportPerson.objects.get(id=player.pk)
        assert person.get_player() == player


@pytest.mark.django_db
class TestPlayer:
    def test_str_full(self):
        player = PlayerFactory(first_name="Test", last_name="Name", current_team=None)
        assert player.str_full() == "Test Name"

    def test_str_full_team(self):
        player = PlayerFactory(current_team__name="Test Team", position=None, first_name="Test", last_name="Name")
        assert player.str_full() == "Test Name (Test Team)"

    def test_str_full_team_position(self):
        player = PlayerFactory(
            current_team__name="Test Team", position="Test Position", first_name="Test", last_name="Name"
        )
        assert player.str_full() == "Test Name (Test Position at Test Team)"


@pytest.mark.django_db
class TestRole:
    def test_str(self):
        role = RoleFactory(title="Test Role")
        assert role.__str__() == "Test Role"


@pytest.mark.django_db
class TestSport:
    def test_str(self):
        sport = SportFactory(name="Test Sport")
        assert sport.__str__() == "Test Sport"

    def test_post_save(self):
        sport = SportFactory()
        league1 = LeagueFactory(sport=sport)
        old_name = sport.name
        new_name = "New Sport Name"
        sport.name = new_name
        sport.save()
        new_computed_str = League.objects.get(id=league1.pk).computed_str
        assert old_name not in new_computed_str
        assert new_name in new_computed_str


@pytest.mark.django_db
class TestLeague:
    def test_str(self):
        league = LeagueFactory(name="Test League", sport__name="Test Sport")
        assert league.__str__() == "Test League (Test Sport)"

    def test_unique_constraint(self):
        league = LeagueFactory()
        with pytest.raises(IntegrityError):
            LeagueFactory(name=league.name, sport=league.sport)

    def test_post_save(self):
        commissioner = SportPersonFactory()
        league = LeagueFactory(commissioner=commissioner)
        old_name = league.name
        new_name = "New League Name"
        league.name = new_name
        league.save()
        new_computed_str = SportPerson.objects.get(id=commissioner.pk).computed_str
        assert old_name not in new_computed_str
        assert new_name in new_computed_str


@pytest.mark.django_db
class TestTeamResults:
    def test_str(self):
        team_results = TeamResultsFactory(team__name="Test Team", league__name="Test League")
        assert team_results.__str__() == "Test Team in Test League"

    @pytest.mark.parametrize("points_for, points_against, difference", [(33, 91, -58), (252, 171, 81), (86, 86, 0)])
    def test_save(self, points_for, points_against, difference):
        team_results = TeamResultsFactory(match_points_for=points_for, match_points_against=points_against)
        assert team_results.match_point_difference == difference

    def test_unique_constraint(self):
        team_results = TeamResultsFactory()
        with pytest.raises(IntegrityError):
            TeamResultsFactory(team=team_results.team, league=team_results.league)


@pytest.mark.django_db
class TestStadium:
    def test_str(self):
        stadium = StadiumFactory(name="Test Stadium")
        assert stadium.__str__() == "Test Stadium"


@pytest.mark.django_db
class TestTeam:
    def test_str(self):
        team = TeamFactory(name="Test Team")
        assert team.__str__() == "Test Team"

    def test_post_save(self):
        team = TeamFactory()
        old_name = team.name
        player1 = PlayerFactory(current_team=team)
        player2 = PlayerFactory(current_team=team)
        player3 = PlayerFactory(current_team=team)
        match1 = MatchFactory(home=team)
        match2 = MatchFactory(away=team)
        new_name = "New Team Name"
        team.name = new_name
        team.save()
        for match in [match1, match2]:
            match_new_computed_str = Match.objects.get(id=match.pk).computed_str
            assert old_name not in match_new_computed_str
            assert new_name in match_new_computed_str
        for player in [player1, player2, player3]:
            player_new_computed_str = Player.objects.get(id=player.pk).computed_str
            assert old_name not in player_new_computed_str
            assert new_name in player_new_computed_str
        coach_new_computed_str = SportPerson.objects.get(id=team.coach.pk).computed_str
        assert old_name not in coach_new_computed_str
        assert new_name in coach_new_computed_str



from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

from celery import shared_task
from django.db import models
from django.db.models.signals import ModelSignal
from langchain_core.language_models import BaseChatModel
from langchain_openai import ChatOpenAI

from ..exceptions import APIStatusErrors
from .utils import construct_prompt, run_llm

if TYPE_CHECKING:
    from pydantic import BaseModel


@shared_task(
    autoretry_for=tuple(APIStatusErrors),
    retry_backoff=10,
    max_retries=5,  # retry i5 times maximum
    default_retry_delay=30,  # retry in 30s
    retry_jitter=False,
)
def invoke_as_task(
    instance,
    prompt: str,
    chat_model: type[BaseChatModel],
    chat_model_name: str,
    max_tokens: int,
    model_field: str | None,
    output_model: "type[BaseModel] | None",
    llm_kwargs: dict[str, Any],
):
    result = run_llm(prompt, instance, output_model, chat_model, chat_model_name, max_tokens, **llm_kwargs)
    if model_field is not None:
        setattr(instance, model_field, result)

    if output_model is not None:
        for field, value in result.items():
            setattr(instance, field, value)
    return instance


T = TypeVar("T", bound=models.Model)


add_llm_prompt = ModelSignal()


class LLMConfig(Generic[T]):
    def __init__(
        self,
        key: str,
        prompt: Callable | str,
        field: str | None = None,
        output_model: "type[BaseModel] | None" = None,
        on_save: bool = True,
        on_condition: Callable | bool | None = None,
        chat_model: type[BaseChatModel] = ChatOpenAI,
        chat_model_name: str = "gpt-4o-mini",
        max_tokens: int = 16000,
        **kwargs
    ):
        self.key = key
        self.on_save = on_save
        self.field = field
        self.on_condition = on_condition
        self.prompt = prompt
        self.output_model = output_model
        self.chat_model = chat_model
        self.chat_model_name = chat_model_name
        self.max_tokens = max_tokens
        self.kwargs = kwargs

    def check_condition(self, instance: T) -> bool:
        if self.on_condition is None:
            return True
        if callable(self.on_condition):
            return self.on_condition(instance)
        return bool(self.on_condition)

    def get_prompt(self, instance: T):
        prompt = construct_prompt(instance, self.prompt)
        for _, response in add_llm_prompt.send(sender=instance.__class__, instance=instance, key=self.key):
            prompt.extend(response)

        return prompt

    def schedule(self, instance: T, initial: bool = True):
        prompt = self.get_prompt(instance)
        args = []
        if initial:
            args.append(instance)
        args.extend(
            [
                prompt,
                self.chat_model,
                self.chat_model_name,
                self.max_tokens,
                self.field,
                self.output_model,
                self.kwargs,
            ]
        )
        return invoke_as_task.s(*args)

    def invoke(self, instance: T):
        self.schedule(instance)()



from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class StadiumDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="city", label=_("City")),
                dp.Field(key="guest_rating", label=_("Rating")),
                dp.Field(
                    key="capacity",
                    label=_("Capacity"),
                    open_by_default=False,
                    children=[
                        dp.Field(key="total_capacity", label=_("Total Capacity")),
                        dp.Field(key="seating_capacity", label=_("Seated"), show="open"),
                        dp.Field(key="standing_capacity", label=_("Standing"), show="open"),
                    ],
                ),
            )
        )

    def get_instance_display(self) -> Display:
        matches_section = Section(
            key="matches_section",
            collapsible=False,
            title=_("Recent Matches"),
            display=Display(
                pages=[
                    Page(
                        title=_("Recent Matches"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["matches_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="matches_inline", endpoint="recent_matches")],
                            )
                        },
                    ),
                ]
            ),
        )
        teams_section = Section(
            key="teams_section",
            title=_("Teams"),
            display=Display(
                pages=[
                    Page(
                        title=_("Teams"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["teams_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="teams_inline", endpoint="teams_playing")],
                            )
                        },
                    ),
                ]
            ),
        )

        is_create_display = "pk" not in self.view.kwargs

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=(
                                [
                                    ["name", "city", "matches_section", "matches_section"],
                                    ["capacity", ".", "matches_section", "matches_section"],
                                    ["guest_rating", ".", "matches_section", "matches_section"],
                                    ["teams_section", "teams_section", "teams_section", "teams_section"],
                                ]
                                if not is_create_display
                                else [["name", "city", "capacity"]]
                            ),
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                            sections=[teams_section, matches_section],
                        ),
                        lte(1100): Layout(
                            grid_template_areas=(
                                [
                                    ["name", "city", "capacity"],
                                    ["matches_section", "matches_section", "matches_section"],
                                    ["teams_section", "teams_section", "teams_section"],
                                    ["guest_rating", ".", "."],
                                ]
                                if not is_create_display
                                else [["name", "city", "capacity"]]
                            ),
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            row_gap=Style.rem(3),
                            sections=[teams_section, matches_section],
                        ),
                        lte(600): Layout(
                            grid_template_areas=(
                                [
                                    ["name", "city"],
                                    ["capacity", "."],
                                    ["matches_section", "matches_section"],
                                    ["teams_section", "teams_section"],
                                    ["guest_rating", "."],
                                ]
                                if not is_create_display
                                else [["name", "city"], ["capacity", "."]]
                            ),
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            sections=[teams_section, matches_section],
                        ),
                        lte(430): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["city"],
                                ["capacity"],
                                ["guest_rating"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    }
                )
            ]
        )



from django.apps import apps
from django.db import connection
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register

from ..factories import ContinentFactory
from .signals import app_pre_migration

register(ContinentFactory)


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))



import pytest
from django.core.cache import cache
from dynamic_preferences.models import global_preferences_registry
from faker import Faker
from wbcore.cache.registry import CachedClass
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.viewsets import ChartViewSet

fake = Faker()


class TestClass(ChartViewSet):
    test = fake.word()

    def _get_dataframe(self):
        return {"figure": "b"}

    def _get_cache_key(self) -> str:
        return "test-class-cache"


@pytest.mark.django_db
class TestCachedClass:
    @pytest.fixture()
    def system_user(self):
        return UserFactory.create(
            is_superuser=True, email=global_preferences_registry.manager()["wbcore__system_user_email"]
        )

    @pytest.fixture()
    def cache_class(self, view_kwargs, get_parameters):
        return CachedClass(
            view_class=TestClass,
            view_kwargs=view_kwargs,
            get_parameters=get_parameters,
        )

    @pytest.mark.parametrize("get_parameters, view_kwargs", [([{"a": "b"}], [{}])])
    def test__get_requests(self, get_parameters, view_kwargs, system_user, cache_class):
        request = list(cache_class._get_requests())[0]
        assert request.user == system_user
        assert request.GET.dict() == get_parameters[0]

    @pytest.mark.parametrize("get_parameters, view_kwargs", [(lambda: [{"a": TestClass.test}], [{}])])
    def test__get_requests_callable(self, system_user, cache_class):
        request = list(cache_class._get_requests())[0]
        assert request.user == system_user
        assert request.GET.dict() == {"a": cache_class.view_class.test}

    @pytest.mark.parametrize("get_parameters, view_kwargs", [([{}], [{"a": "b"}])])
    def test_fetch_cache(self, system_user, cache_class):
        cache.set(
            "test-class-cache", {"other_figure": "c"}
        )  # ensure that fetch cache clear cache before refetching result

        res = cache_class.fetch_cache()[0]
        assert res == {"figure": "b"}



from wbcore.configs.registry import ConfigRegistry


def test_workflow_config(config_registry: ConfigRegistry):
    workflow = config_registry.get_config_dict()["workflow"]
    assert workflow["endpoint"]



from typing import Any, Iterable

from rest_framework import serializers

from .mixins import WBCoreSerializerFieldMixin
from .types import WBCoreType


class BooleanField(WBCoreSerializerFieldMixin, serializers.BooleanField):
    field_type = WBCoreType.BOOLEAN.value

    def __init__(
        self,
        labels: Iterable[str] | None = None,
        values: Iterable[Any] | None = None,
        background_color: str | Iterable[str] | None = None,
        color: str | Iterable[str] | None = None,
        active_background_color: str | Iterable[str] | None = None,
        active_color: str | Iterable[str] | None = None,
        **kwargs
    ):
        self.labels = labels
        self.values = values
        self.background_color = background_color
        self.color = color
        self.active_background_color = active_background_color
        self.active_color = active_color
        super().__init__(**kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        field_name, representation = super().get_representation(request, field_name)

        if self.labels:
            representation["labels"] = self.labels

        if self.values:
            representation["values"] = self.values

        if self.background_color:
            representation["background_color"] = self.background_color

        if self.color:
            representation["color"] = self.color

        if self.active_background_color:
            representation["active_background_color"] = self.active_background_color

        if self.active_color:
            representation["active_color"] = self.active_color

        return field_name, representation



from enum import Enum
from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import ClientManagerRelationship as CMR
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class CMR_Color(Enum):
    DRAFT = WBColor.RED_LIGHT.value
    PENDING = WBColor.YELLOW_LIGHT.value
    APPROVED = WBColor.GREEN_LIGHT.value
    PENDING_REMOVE = WBColor.YELLOW_DARK.value
    REMOVED = WBColor.GREY.value


class RelationshipDisplayConfig(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display([["from_entry", "to_entry", "relationship_type"]])

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="relationship_type", label=_("Type")),
                dp.Field(key="from_entry", label=_("From")),
                dp.Field(key="to_entry", label=_("To")),
            ]
        )


class RelationshipEntryDisplay(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display([["relationship_type", "to_entry"]])

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="relationship_type", label=_("Relationship Type")),
                dp.Field(key="to_entry", label=_("Relationship To")),
            ]
        )


class RelationshipTypeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label=_("Name")),
                dp.Field(key="counter_relationship", label=_("Counter Relationship")),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["title", "counter_relationship"]])


class EmployerEmployeeRelationshipDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="employer", label=_("Employer")),
                dp.Field(key="employee", label=_("Employee")),
                dp.Field(key="primary", label=_("Company Is Primary Employer")),
                dp.Field(key="position", label=_("Position In Company")),
            ],
            legends=[
                dp.Legend(
                    key="primary",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Company Is Primary Employer"),
                            value=True,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                ),
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(icon=WBIcon.FAVORITE.icon, condition=("==", True)),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["employer", "employee"],
                ["primary", "position"],
            ]
        )


class EmployeeEmployerDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="employee", label=_("Employee")),
                dp.Field(key="primary", label=_("Company Is Primary Employer")),
                dp.Field(key="position", label=_("Position In Company")),
                dp.Field(key="position_name", label=_("Position Name")),
                dp.Field(key="employee_profile_pic", label=_("Profile Picture")),
            ],
            legends=[
                dp.Legend(
                    key="primary",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Company Is Primary Employer"),
                            value=True,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                ),
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(icon=WBIcon.FAVORITE.icon, condition=("==", True)),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([[repeat_field(2, "employee")], ["primary", "position"]])


class EmployerEmployeeDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="employer", label=_("Employer")),
                dp.Field(key="primary", label=_("Company Is Primary Employer")),
                dp.Field(key="position", label=_("Position In Company")),
                dp.Field(key="position_name", label=_("Position Name")),
                dp.Field(key="employer_profile_pic", label=_("Profile Picture")),
            ],
            legends=[
                dp.Legend(
                    key="primary",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Company Is Primary Employer"),
                            value=True,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                ),
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(icon=WBIcon.FAVORITE.icon, condition=("==", True)),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([[repeat_field(2, "employer")], ["primary", "position"]])


class ClientManagerModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="client", label=_("Client")),
            dp.Field(key="relationship_manager", label=_("Relationship Manager")),
            dp.Field(key="created", label=_("Created")),
        ]
        if "client" in self.request.GET:
            fields = fields[1:]
        elif "relationship_manager" in self.request.GET:
            fields = fields[0::2]

        return dp.ListDisplay(
            fields=fields,
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=CMR_Color.DRAFT.value,
                            label=CMR.Status.DRAFT.label,
                            value=CMR.Status.DRAFT.name,
                        ),
                        dp.LegendItem(
                            icon=CMR_Color.PENDING.value,
                            label=CMR.Status.PENDINGADD.label,
                            value=CMR.Status.PENDINGADD.name,
                        ),
                        dp.LegendItem(
                            icon=CMR_Color.APPROVED.value,
                            label=CMR.Status.APPROVED.label,
                            value=CMR.Status.APPROVED.name,
                        ),
                        dp.LegendItem(
                            icon=CMR_Color.PENDING_REMOVE.value,
                            label=CMR.Status.PENDINGREMOVE.label,
                            value=CMR.Status.PENDINGREMOVE.name,
                        ),
                        dp.LegendItem(
                            icon=CMR_Color.REMOVED.value,
                            label=CMR.Status.REMOVED.label,
                            value=CMR.Status.REMOVED.name,
                        ),
                    ],
                ),
                dp.Legend(
                    key="primary",
                    items=[
                        dp.LegendItem(icon=WBIcon.FAVORITE.icon, label=_("Primary"), value=True),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", CMR.Status.PENDINGADD.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_DARK.value},
                            condition=("==", CMR.Status.PENDINGREMOVE.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", CMR.Status.APPROVED.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", CMR.Status.DRAFT.name),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREY.value},
                            condition=("==", CMR.Status.REMOVED.name),
                        ),
                    ],
                ),
                dp.Formatting(
                    column="primary",
                    formatting_rules=[
                        dp.FormattingRule(icon=WBIcon.FAVORITE.icon, condition=("==", True)),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        fields = []
        if "pk" not in self.view.kwargs:
            fields.append([repeat_field(2, "primary")])

        else:
            fields.append(["status", "primary"])

        if "client" in self.request.GET:
            fields.append(["relationship_manager", "."])
        else:
            fields.append(["client", "relationship_manager"])

        return create_simple_display(fields)


class UserIsClientDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="relationship_manager_name", label=_("Name")),
                dp.Field(key="relationship_manager_email", label=_("Telephone")),
                dp.Field(key="relationship_manager_phone_number", label=_("Email")),
            ]
        )



from unittest.mock import patch

import pytest
from wbcore.contrib.workflow.models import ProcessStep


@pytest.mark.django_db
class TestJoinStep:
    @patch("wbcore.contrib.workflow.models.step.Step.set_canceled")
    def test_cancel_if_leading_to_self_transition_found(
        self, mock_canceled, random_child_step_factory, join_step_factory, process_step_factory, transition_factory
    ):
        self_step = join_step_factory()
        step = random_child_step_factory()
        process_step = process_step_factory(step=step)
        transition1 = transition_factory(from_step=step)
        transition_factory(from_step=step)
        transition_factory(from_step=transition1.to_step, to_step=self_step)
        transition_factory()
        self_step.cancel_if_leading_to_self(step, process_step)
        assert mock_canceled.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.Step.set_canceled")
    def test_cancel_if_leading_to_self_no_transition_found(
        self,
        mock_canceled,
        random_child_step_factory,
        join_step_factory,
        process_step_factory,
        transition_factory,
    ):
        self_step = join_step_factory()
        step = random_child_step_factory()
        process_step = process_step_factory(step=step)
        transition1 = transition_factory(from_step=step)
        transition_factory(from_step=step)
        transition_factory(from_step=transition1.to_step)
        transition_factory()
        self_step.cancel_if_leading_to_self(step, process_step)
        assert not mock_canceled.called

    @patch("wbcore.contrib.workflow.models.step.JoinStep.cancel_if_leading_to_self")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run_wait_for_all_transition_without_process_step(
        self, mock_execute, mock_cancel, process_step_factory, join_step_factory, transition_factory
    ):
        step = join_step_factory(wait_for_all=True)
        transition_factory(to_step=step)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        step.run(process_step)
        assert process_step.state == ProcessStep.StepState.WAITING
        assert not mock_execute.called
        assert not mock_cancel.called

    @patch("wbcore.contrib.workflow.models.step.JoinStep.cancel_if_leading_to_self")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run_wait_for_all_transition_with_process_step(
        self, mock_execute, mock_cancel, process_step_factory, join_step_factory, transition_factory
    ):
        step = join_step_factory(wait_for_all=True)
        transition = transition_factory(to_step=step)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        process_step_factory(
            step=transition.from_step, process=process_step.process, state=ProcessStep.StepState.ACTIVE
        )
        step.run(process_step)
        assert process_step.state == ProcessStep.StepState.WAITING
        assert not mock_execute.called
        assert not mock_cancel.called

    @pytest.mark.parametrize("state", [ProcessStep.StepState.FINISHED, ProcessStep.StepState.CANCELED])
    @patch("wbcore.contrib.workflow.models.step.JoinStep.cancel_if_leading_to_self")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run_wait_for_all_transition_with_finished_canceled_process_step(
        self, mock_execute, mock_cancel, process_step_factory, join_step_factory, transition_factory, state
    ):
        step = join_step_factory(wait_for_all=True)
        transition = transition_factory(to_step=step)
        transition_factory(from_step=step)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        process_step_factory(step=transition.from_step, process=process_step.process, state=state)
        step.run(process_step)
        assert process_step.state == ProcessStep.StepState.WAITING
        assert mock_execute.call_args.args == (process_step,)
        assert not mock_cancel.called

    @patch("wbcore.contrib.workflow.models.step.JoinStep.cancel_if_leading_to_self")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run_wait_for_all_no_transition(
        self, mock_execute, mock_cancel, process_step_factory, join_step_factory, transition_factory
    ):
        step = join_step_factory(wait_for_all=True)
        transition_factory(from_step=step)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        step.run(process_step)
        assert process_step.state == ProcessStep.StepState.WAITING
        assert mock_execute.call_args.args == (process_step,)
        assert not mock_cancel.called

    @patch("wbcore.contrib.workflow.models.step.JoinStep.cancel_if_leading_to_self")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run_do_not_wait(
        self,
        mock_execute,
        mock_cancel,
        process_step_factory,
        join_step_factory,
    ):
        step = join_step_factory(wait_for_all=False)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        unfinished_process_step1 = process_step_factory(
            process=process_step.process, state=ProcessStep.StepState.ACTIVE
        )
        unfinished_process_step2 = process_step_factory(
            process=process_step.process, state=ProcessStep.StepState.WAITING
        )
        process_step_factory(state=ProcessStep.StepState.ACTIVE)
        process_step_factory(process=process_step.process, state=ProcessStep.StepState.FINISHED)
        process_step_factory(process=process_step.process, state=ProcessStep.StepState.CANCELED)
        process_step_factory(process=process_step.process, state=ProcessStep.StepState.FAILED)
        step.run(process_step)
        assert process_step.state == ProcessStep.StepState.WAITING
        assert mock_execute.call_args.args == (process_step,)
        assert set(tuple(map(lambda y: str(y.pk), x.args)) for x in mock_cancel.call_args_list) == {
            (str(unfinished_process_step1.step.pk), unfinished_process_step1.pk),
            (str(unfinished_process_step2.step.pk), unfinished_process_step2.pk),
        }



from django.db import models
from django.utils.translation import gettext_lazy as _
from wbcore.contrib.agenda.typings import ConferenceRoom as ConferenceRoomDTO
from wbcore.models import WBModel


class Building(WBModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    address = models.ForeignKey(
        on_delete=models.SET_NULL,
        to="directory.AddressContact",
        verbose_name=_("Address"),
        related_name="conference_room_building",
        null=True,
        blank=True,
    )

    def __str__(self) -> str:
        return f"{self.name}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:agenda:building"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:agenda:buildingrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    class Meta:
        verbose_name = _("Building")
        constraints = [
            models.UniqueConstraint(
                fields=["name", "address"],
                name="unique_building",
            ),
        ]


class ConferenceRoom(WBModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    email = models.EmailField(verbose_name=_("Email Address"), unique=True)
    building = models.ForeignKey(
        on_delete=models.CASCADE,
        to=Building,
        verbose_name=_("Building"),
        related_name="conference_room",
    )
    capacity = models.PositiveIntegerField(verbose_name=_("Capacity"), null=True, blank=True)
    is_videoconference_capable = models.BooleanField(verbose_name=_("Capable of Videoconferencing"), default=True)

    def __str__(self) -> str:
        return f"{self.name} ({self.building.name})"

    def _build_dto(self):
        return ConferenceRoomDTO(
            name=self.name,
            email=self.email,
            name_building=self.building.name,
            capacity=self.capacity,
            is_videoconference_capable=self.is_videoconference_capable,
            id=self.id,
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:agenda:conferenceroom"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:agenda:conferenceroomrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}} ({{_building.name}})"

    class Meta:
        verbose_name = _("Conference Room")
        constraints = [
            models.UniqueConstraint(
                fields=["name", "building"],
                name="unique_conference_room",
            ),
        ]



from langchain_core.exceptions import LangChainException

APIStatusErrors = [LangChainException]

try:
    from openai._exceptions import (
        AuthenticationError,
        BadRequestError,
        ConflictError,
        InternalServerError,
        NotFoundError,
        PermissionDeniedError,
        RateLimitError,
        UnprocessableEntityError,
    )

    APIStatusErrors.extend(
        [
            BadRequestError,
            AuthenticationError,
            PermissionDeniedError,
            NotFoundError,
            ConflictError,
            UnprocessableEntityError,
            RateLimitError,
            InternalServerError,
        ]
    )
except ImportError:
    pass


try:
    from anthropic._exceptions import (
        APIConnectionError,
        APIResponseValidationError,
        APIStatusError,
    )

    APIStatusErrors.extend([APIResponseValidationError, APIStatusError, APIConnectionError])
except ImportError:
    pass



import factory
from wbcore.contrib.notifications.models import (
    NotificationType,
    NotificationTypeSetting,
)


class NotificationTypeModelFactory(factory.django.DjangoModelFactory):
    code = factory.Faker("pystr")
    title = factory.Faker("pystr")
    help_text = factory.Faker("pystr")

    class Meta:
        model = NotificationType


class NotificationTypeSettingModelFactory(factory.django.DjangoModelFactory):
    notification_type = factory.SubFactory(NotificationTypeModelFactory)
    user = factory.SubFactory("wbcore.contrib.authentication.factories.UserFactory")

    class Meta:
        model = NotificationTypeSetting



from wbcore.metadata.configs.display.instance_display.shortcuts import Display

from .base import WBCoreViewConfig


class PreviewViewConfig(WBCoreViewConfig):
    metadata_key = "preview"
    config_class_attribute = "preview_config_class"

    DISPLAY_TYPE = "instance_display"

    def get_buttons(self):
        return []

    def _get_buttons(self):
        return [dict(button) for button in self.get_buttons()]

    def get_display(self) -> Display | None:
        return None

    def _get_display(self):
        if display := self.get_display():
            if self.DISPLAY_TYPE == "instance_display":
                return display.serialize()
            return display
        return None

    def _get_display_type(self):
        return self.DISPLAY_TYPE

    def get_metadata(self) -> dict:
        buttons = self._get_buttons()
        display = self._get_display()

        preview = dict()
        if len(buttons) > 0 or display is not None:
            preview["buttons"] = buttons
            preview["display"] = display
            preview["type"] = self._get_display_type()

        return preview



from django.contrib.postgres.search import TrigramSimilarity
from django.db.models.expressions import Value
from rest_framework.filters import OrderingFilter
from wbcore import viewsets
from wbcore.contrib.guardian.filters import ObjectPermissionsFilter
from wbcore.viewsets.mixins import DjangoFilterBackend

from ..models import Geography
from ..serializers import GeographyModelSerializer, GeographyRepresentationSerializer
from .display import GeographyDisplayConfig
from .preview import GeographyPreviewConfig
from .titles.geography import GeographyTitleConfig


class GeographyRepresentationViewSet(viewsets.RepresentationViewSet):
    filter_backends = (ObjectPermissionsFilter, DjangoFilterBackend, OrderingFilter)
    ordering = ["-trigram", "ranking"]
    filterset_fields = {"level": ["exact"]}

    queryset = Geography.objects.all()
    serializer_class = GeographyRepresentationSerializer

    def get_queryset(self):
        queryset = super().get_queryset()

        if search := self.request.GET.get("search", None):
            if search.startswith("^"):
                return queryset.annotate(trigram=Value(1)).filter(search_vector=search[1:])

            return queryset.filter(trigram_search_vector__icontains=search).annotate(
                trigram=TrigramSimilarity("name", search)
            )

        return queryset.annotate(trigram=Value(1))


class GeographyModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "geography/markdown/documentation/geography.md"
    display_config_class = GeographyDisplayConfig
    preview_config_class = GeographyPreviewConfig
    title_config_class = GeographyTitleConfig
    ordering_fields = [
        "code_2",
        "code_3",
        "parent__name",
    ]
    ordering = ["name"]
    search_fields = ["name"]
    filterset_fields = {
        "code_2": ["exact", "icontains"],
        "code_3": ["exact", "icontains"],
        "parent": ["exact"],
        "level": ["exact"],
    }

    queryset = Geography.objects.all()
    serializer_class = GeographyModelSerializer



from django.contrib import admin
from wbcore.contrib.agenda.models import CalendarItem


@admin.register(CalendarItem)
class CalendarItemModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("entities",)
    list_display = ["item_type", "title", "period", "is_active"]
    search_fields = ("entities__computed_str",)
    raw_id_fields = ["entities"]

    def get_queryset(self, request):
        return CalendarItem.all_objects.all()



from wbcore import serializers

from .models import ReleaseNote


class ReleaseNoteModelSerializer(serializers.ModelSerializer):
    user_read = serializers.BooleanField()

    class Meta:
        model = ReleaseNote
        fields = (
            "id",
            "version",
            "release_date",
            "module",
            "summary",
            "notes",
            "user_read",
        )



from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import Q, UniqueConstraint
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from slugify import slugify
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin, PrimaryMixin


class RelationshipType(WBModel):
    title = models.CharField(
        max_length=256,
        verbose_name=_("Relationship Type"),
        unique=True,
        blank=False,
        null=False,
    )

    counter_relationship = models.OneToOneField(
        to="self",
        blank=True,
        null=True,
        related_name="reverse_counter_relationship",
        on_delete=models.CASCADE,
        verbose_name=_("Counter Relationship"),
    )

    slugify_title = models.CharField(
        max_length=256,
        unique=True,
        blank=True,
        null=True,
        verbose_name="Slugified Title",
    )

    def __str__(self) -> str:
        return f"{self.title}"

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        if not self.id:
            # self will have no id, when it is getting creating. In this case post_save will update
            # the counter relationship if one is given.
            super().save(*args, **kwargs)
        else:
            if self.counter_relationship:
                RelationshipType.objects.filter(
                    Q(reverse_counter_relationship=self) & ~Q(id=self.counter_relationship.id)
                ).update(counter_relationship=None)

                RelationshipType.objects.filter(id=self.counter_relationship.id).update(counter_relationship=self)
            else:
                RelationshipType.objects.filter(reverse_counter_relationship=self).update(counter_relationship=None)

            super().save(*args, **kwargs)

    class Meta:
        verbose_name = _("Relationship Type")
        verbose_name_plural = _("Relationship Types")

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:relationship-type"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:relationshiptyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"


class Relationship(ComplexToStringMixin, WBModel):
    relationship_type = models.ForeignKey(
        on_delete=models.PROTECT,
        to="directory.RelationshipType",
        verbose_name=_("Type"),
    )
    from_entry = models.ForeignKey(
        on_delete=models.CASCADE,
        null=True,
        related_name="from_entry",
        to="directory.Entry",
        verbose_name=_("Relationship From"),
    )
    to_entry = models.ForeignKey(
        on_delete=models.CASCADE,
        null=True,
        related_name="to_entry",
        to="directory.Entry",
        verbose_name=_("Relationship To"),
    )

    def compute_str(self) -> str:
        return f"{self.from_entry} is {self.relationship_type} of {self.to_entry}"

    def delete(self, *args, **kwargs):
        Relationship.objects.filter(
            Q(to_entry=self.from_entry)
            & Q(from_entry=self.to_entry)
            & (
                Q(relationship_type=self.relationship_type)
                | Q(relationship_type=self.relationship_type.counter_relationship)
            )
        ).delete()
        super().delete(*args, **kwargs)

    class Meta:
        verbose_name = _("Relationship")
        verbose_name_plural = _("Relationships")

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:relationship"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:relationshiprepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{relationship_type}}: {{from_entry}} {{to_entry}}"


class ClientManagerRelationship(ComplexToStringMixin, PrimaryMixin, WBModel):
    class Status(models.TextChoices):
        DRAFT = "DRAFT", _("Draft")
        PENDINGADD = "PENDINGADD", _("Pending Add")
        PENDINGREMOVE = "PENDINGREMOVE", _("Pending Remove")
        APPROVED = "APPROVED", _("Approved")
        REMOVED = "REMOVED", _("Removed")

    PRIMARY_ATTR_FIELDS = ["client"]

    relationship_manager = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Person",
        verbose_name=_("Relationship Manager"),
        related_name="manager_of",
    )
    primary = models.BooleanField(
        verbose_name=_("Primary"),
        default=False,
    )
    client = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Entry",
        verbose_name=_("Client"),
        related_name="client_of",
    )
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name=_("Status"))

    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))

    @transition(
        field=status,
        source=[Status.DRAFT],
        target=Status.PENDINGADD,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.PRIMARY,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label=_("Submit"),
                action_label=_("Submitting"),
                description_fields=_("Are you sure you want to submit your draft for review?"),
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.DRAFT],
        target=Status.APPROVED,
        permission=lambda instance, user: user.has_perm("directory.change_clientmanagerrelationship"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.SUCCESS,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.CONFIRM.icon,
                key="mngapprove",
                label=_("Approve"),
                action_label=_("Approve"),
                description_fields=_("Are you sure you want to approve this relationship?"),
            )
        },
    )
    def mngapprove(self, by=None, description=None, **kwargs):
        if (
            self.primary is True
            and ClientManagerRelationship.objects.filter(client=self.client, primary=True).exists()
        ):
            ClientManagerRelationship.objects.filter(client=self.client, primary=True).update(primary=False)

        if (
            self.primary is False
            and not ClientManagerRelationship.objects.filter(client=self.client, primary=True).exists()
        ):
            self.primary = True

    @transition(
        field=status,
        source=[Status.PENDINGADD],
        target=Status.DRAFT,
        permission=lambda instance, user: user.has_perm("directory.change_clientmanagerrelationship"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.ERROR,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.REJECT.icon,
                key="deny",
                label=_("Deny"),
                action_label=_("Denial"),
                description_fields=_("Are you sure you want to deny this relationship?"),
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.PENDINGADD],
        target=Status.APPROVED,
        permission=lambda instance, user: user.has_perm("directory.change_clientmanagerrelationship"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.CONFIRM.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="approve",
                label=_("Approve"),
                action_label=_("Approval"),
                description_fields=_("Are you sure you want to approve this relationship?"),
            )
        },
    )
    def approve(self, by=None, description=None, **kwargs):
        if (
            self.primary is True
            and ClientManagerRelationship.objects.filter(client=self.client, primary=True).exists()
        ):
            ClientManagerRelationship.objects.filter(client=self.client, primary=True).update(primary=False)

        if (
            self.primary is False
            and not ClientManagerRelationship.objects.filter(client=self.client, primary=True).exists()
        ):
            self.primary = True

    @transition(
        field=status,
        source=[Status.PENDINGREMOVE],
        target=Status.APPROVED,
        permission=lambda instance, user: user.has_perm("directory.change_clientmanagerrelationship"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.REJECT.icon,
                color=ButtonDefaultColor.ERROR,
                key="denyremoval",
                label=_("Deny"),
                action_label=_("Denial"),
                description_fields=_("Are you sure you want to deny removal of this relationship?"),
            )
        },
    )
    def denyremoval(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.PENDINGREMOVE],
        target=Status.REMOVED,
        permission=lambda instance, user: user.has_perm("directory.change_clientmanagerrelationship"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.DELETE.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="approveremoval",
                label=_("Approve"),
                action_label=_("Approval"),
                description_fields=_("Are you sure you want to remove this relationship?"),
            )
        },
    )
    def approveremoval(self, by=None, description=None, **kwargs):
        pass

    def is_not_primary(instance):
        return not instance.primary

    @transition(
        field=status,
        source=[Status.APPROVED],
        target=Status.PENDINGADD,
        conditions=[is_not_primary],
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.FAVORITE.icon,
                color=ButtonDefaultColor.PRIMARY,
                key="makeprimary",
                label=_("Make Primary"),
                action_label=_("Make Primary"),
                description_fields=_("Are you sure you want to request making this relationship manager primary?"),
            )
        },
    )
    def makeprimary(self, by=None, description=None, **kwargs):
        self.primary = True

    def last_primary(instance):
        return not instance.primary and (
            ClientManagerRelationship.objects.exclude(id=instance.id)
            .filter(status=ClientManagerRelationship.Status.APPROVED, client=instance.client, primary=True)
            .exists()
        )

    @transition(
        field=status,
        source=[Status.APPROVED],
        target=Status.PENDINGREMOVE,
        conditions=[last_primary],
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.DELETE.icon,
                color=ButtonDefaultColor.ERROR,
                key="remove",
                label=_("Remove"),
                action_label=_("Removal"),
                description_fields=_("Are you sure you want to request removal of this relationship?"),
            )
        },
    )
    def remove(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.REMOVED],
        target=Status.PENDINGADD,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("directory:clientmanagerrelationship",),
                icon=WBIcon.UNDO.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="reinstate",
                label=_("Reinstate"),
                action_label=_("Reinstate"),
                description_fields=_("Are you sure you want to request reinstating this relationship?"),
            )
        },
    )
    def reinstate(self, by=None, description=None, **kwargs):
        pass

    def compute_str(self) -> str:
        return _("{client} is client of {manager}").format(client=self.client, manager=self.relationship_manager)

    def delete(self, **kwargs):
        super().delete(no_deletion=False)  # For this model we actually want to delete the object

    def get_related_queryset(self):
        return (
            super().get_related_queryset().filter(status=self.Status.APPROVED)
        )  # only one approved relationship can be primary

    class Meta:
        verbose_name = _("Client Manager Relationship")
        verbose_name_plural = _("Client Manager Relationships")

        notification_types = [
            create_notification_type(
                code="directory.clientmanagerrelationship.approval",
                title="Relationship Manager Approval",
                help_text="Sends you a notification when there is a Relationship manager change to approve.",
            )
        ]

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:clientmanagerrelationshiprepresentation-list"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:clientmanagerrelationship"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{relationship_manager}} {{client}} {{primary}}"


@receiver(post_save, sender=ClientManagerRelationship)
def post_save_client_manager_relationship(sender, instance, created, **kwargs):
    """
    Post save signal receiver sending notifications to all managers when a new client manager relationship is requested
    """
    if instance.status in [
        ClientManagerRelationship.Status.PENDINGADD.name,
        ClientManagerRelationship.Status.PENDINGREMOVE.name,
    ]:
        user_qs = (
            get_user_model()
            .objects.filter(
                models.Q(groups__permissions__codename="can_administer_entry_in_charge_request")
                | models.Q(user_permissions__codename="can_administer_entry_in_charge_request")
            )
            .distinct()
        )
        if instance.status == ClientManagerRelationship.Status.PENDINGADD.name:
            for user in user_qs.all():
                send_notification(
                    code="directory.clientmanagerrelationship.approval",
                    title=_("New Client Manager Relationship for {client}").format(
                        client=instance.client.computed_str
                    ),
                    body=_("User requested that {manager} is the relationship manager of {client}.").format(
                        manager=str(instance.relationship_manager), client=instance.client.computed_str
                    ),
                    user=user,
                    reverse_name="directory:clientmanagerrelationship-detail",
                    reverse_args=[instance.id],
                )
        if instance.status == ClientManagerRelationship.Status.PENDINGREMOVE.name:
            for user in user_qs.all():
                send_notification(
                    code="directory.clientmanagerrelationship.approval",
                    title=_("Removal of Client Manager Relationship for {client}").format(
                        client=instance.client.computed_str
                    ),
                    body=_("User requested that relationship manager {manager} of {client} should be removed.").format(
                        manager=str(instance.relationship_manager), client=instance.client.computed_str
                    ),
                    user=user,
                    reverse_name="directory:clientmanagerrelationship-detail",
                    reverse_args=[instance.id],
                )


class Position(WBModel):
    title = models.CharField(max_length=128, verbose_name=_("Title"), unique=True, blank=False, null=False)

    slugify_title = models.CharField(
        max_length=128, unique=True, verbose_name="Slugified Title", blank=True, null=True
    )

    def __str__(self) -> str:
        return f"{self.title}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:position"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:directory:positionrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    class Meta:
        verbose_name = _("Position")
        verbose_name_plural = _("Positions")

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)


class EmployerEmployeeRelationship(PrimaryMixin):
    PRIMARY_ATTR_FIELDS = ["employee"]

    employee = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Person",
        verbose_name=_("Employee"),
        related_name="employer",
    )
    primary = models.BooleanField(
        verbose_name=_("Primary"),
        default=False,
    )
    position = models.ForeignKey(
        null=True,
        on_delete=models.SET_NULL,
        to="directory.Position",
        verbose_name=_("Position"),
        related_name="position_of_employee",
    )
    position_name = models.CharField(
        verbose_name=_("Alternative Position Name"),
        help_text=_(
            "If the selected position isn't an exact match, feel free to use this field to provide more details about the position. You can edit this field once the position is confirmed."
        ),
        max_length=128,
        null=True,
        blank=True,
    )
    employer = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Company",
        verbose_name=_("Employer"),
        related_name="employee",
    )

    def save(self, *args, **kwargs):
        if self.position and not self.position_name:
            self.position_name = str(self.position)
        super().save(*args, **kwargs)

    class Meta:
        constraints = [
            UniqueConstraint(
                name="unique_employeremployee",
                fields=["employee", "employer"],
            )
        ]
        verbose_name = _("Employer Employee Relationship")

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:directory:employeremployeerelationship"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    def representation_label_key(cls):
        return "{{position_name}}"

    def delete(self, **kwargs):
        super().delete(no_deletion=False)  # For this model we actually want to delete the object


@receiver(post_save, sender=EmployerEmployeeRelationship)
def post_save_eer(sender, instance, created, raw, **kwargs):
    """
    Post save EER signal: Triggers the post_save signals of the employee which updates his computed_str and adds the
    employer to future planned activities if it became the only employer
    """

    if not raw:
        instance.employee.save()


@receiver(post_delete, sender=EmployerEmployeeRelationship)
def post_delete_eer(sender, instance, **kwargs):
    """
    Post delete EER signal: Triggers the post_delete signals of the employee which updates his computed_str and adds the
    employer to future planned activities if it became the only employer
    """
    new_computed_str = instance.employee.compute_str()
    instance.employee._meta.model.objects.filter(id=instance.employee.id).update(
        computed_str=new_computed_str, slugify_computed_str=slugify(new_computed_str, separator=" ")
    )


@receiver(post_save, sender=RelationshipType)
def post_save_relationship_type(sender, instance, created, raw, **kwargs):
    if created and not raw:
        if instance.counter_relationship:
            sender.objects.filter(id=instance.counter_relationship.id).update(counter_relationship=instance)


@receiver(post_save, sender=Relationship)
def post_save_relationship(sender, instance, created, raw, **kwargs):
    if raw:
        return
    rel_ship_type = (
        instance.relationship_type.counter_relationship
        if instance.relationship_type.counter_relationship
        else instance.relationship_type
    )
    # The type of relationship the to_entry should have.
    qs = sender.objects.filter(
        from_entry=instance.to_entry, to_entry=instance.from_entry, relationship_type=rel_ship_type
    )

    if not created:
        # An existing instance was changed
        if not qs.exists():
            sender.objects.filter(
                from_entry=instance.to_entry,
                to_entry=instance.from_entry,
            ).update(relationship_type=rel_ship_type)
    else:
        # A new instance was created
        if not qs.exists():
            sender.objects.create(
                from_entry=instance.to_entry,
                to_entry=instance.from_entry,
                relationship_type=rel_ship_type,
            )



import pytest
from faker import Faker
from wbcore.cache.decorators import cache_table
from wbcore.cache.mixins import CacheMixin
from wbcore.utils.strings import camel_to_snake_case

fake = Faker()


class TestCacheMixin:
    @pytest.fixture
    def view_class(self):
        @cache_table(
            timeout=lambda view: view.timeout,
            key_prefix=lambda view: view.get_parameter_value,
        )
        class CacheTable(CacheMixin):
            timeout = fake.pyint()
            get_parameter_value = fake.word()

        return CacheTable

    def test_get_cache_timeout(self, view_class):
        view = view_class()
        assert view._get_cache_timeout() == view_class.timeout

    def test_get_cache_key(self, view_class):
        view = view_class()
        assert view._get_cache_key() == camel_to_snake_case(view_class.__name__) + "-" + view_class.get_parameter_value



import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.directory.models import EmployerEmployeeRelationship, Person

from ..filters.entries import PersonFilter


@pytest.mark.django_db
class TestSpecificFiltersEntries:
    # def test_filter_address(self, address_contact_factory, company_factory):
    #     address_contact_factory(city="Geneva")
    #     mvs = PersonModelViewSet(kwargs={})
    #     mvs.request = APIRequestFactory().get("")
    #     qs = mvs.get_serializer_class().Meta.model.objects.all()
    #     assert mvs.filterset_class().filter_address(qs, "", "Geneva")
    #     assert mvs.filterset_class().filter_address(qs, "", "Lausanne").count() == 0
    #
    #     mvs = CompanyModelViewSet(kwargs={})
    #     mvs.request = APIRequestFactory().get("")
    #     address_contact_factory(city="Geneva", entry=company_factory())
    #     qs = mvs.get_serializer_class().Meta.model.objects.all()
    #     assert mvs.filter_class().filter_address(qs, "", "Geneva")
    #     assert mvs.filter_class().filter_address(qs, "", "Lausanne").count() == 0

    def test_get_union_employee(self, company_factory, user_factory):
        request = APIRequestFactory().get("")
        user1 = user_factory.create(is_active=True, is_superuser=True)
        user2 = user_factory.create(is_active=True, is_superuser=True)
        company1 = company_factory.create()
        company2 = company_factory.create()
        EmployerEmployeeRelationship.objects.create(employee=user1.profile, employer=company1, primary=True)
        EmployerEmployeeRelationship.objects.create(employee=user2.profile, employer=company2, primary=True)

        request.user = user1
        persons = Person.objects.all()
        assert (
            PersonFilter(request=request)
            .get_union_employee(persons, "", company1)
            .filter(id=user1.profile.id)
            .exists()
        )  #
        assert (
            PersonFilter(request=request)
            .get_union_employee(persons, "", company2)
            .filter(id=user1.profile.id)
            .exists()
        )
        assert (
            not PersonFilter(request=request)
            .get_union_employee(persons, "", company1)
            .filter(id=user2.profile.id)
            .exists()
        )  #
        assert (
            PersonFilter(request=request)
            .get_union_employee(persons, "", company2)
            .filter(id=user2.profile.id)
            .exists()
        )



from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from wbcore.shares.signals import handle_widget_sharing
from wbcore.utils.urls import clean_shareable_url

from .sites import share_site


class ShareAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request: Request) -> Response:
        if widget_endpoint := request.POST.get("widget_endpoint"):
            serializer = share_site.serializer_class(data=request.data)
            serializer.is_valid(raise_exception=True)
            handle_widget_sharing.send(
                sender=self.__class__,
                request=request,
                widget_relative_endpoint=clean_shareable_url(widget_endpoint),
                **serializer.validated_data.copy(),
            )
        return Response({})



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.contrib.icons.icons import WBIcon
from wbcore.metadata.configs.buttons.buttons import (
    ActionButton,
    HyperlinkButton,
    RequestType,
    WidgetButton,
)
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display import Field, ListDisplay
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class NotificationButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set[ActionButton]:
        if not self.view.kwargs.get("pk", None):
            return {
                ActionButton(
                    weight=0,
                    method=RequestType.PATCH,
                    action_label=_("Reading all notifications"),
                    endpoint=reverse("wbcore:notifications:notification-read-all", request=self.request),
                    description_fields=_("Do you want to mark notifications as read?"),
                    label=_("Mark all read"),
                    icon=WBIcon.VIEW.icon,
                    identifiers=["notifications:notification"],
                ),
                ActionButton(
                    weight=1,
                    method=RequestType.PATCH,
                    action_label=_("Deleting all read notifications"),
                    endpoint=reverse("wbcore:notifications:notification-delete-all-read", request=self.request),
                    description_fields=_("Do you want delete all read notifications?"),
                    label=_("Delete all read"),
                    icon=WBIcon.DELETE.icon,
                    identifiers=["notifications:notification"],
                ),
            }
        return set()

    def get_custom_instance_buttons(self) -> set[WidgetButton]:
        return {
            WidgetButton(
                title=_("Open Resource"),
                label=_("Open Resource"),
                icon=WBIcon.LINK.icon,
                key="open_internal_resource",
            ),
            HyperlinkButton(
                title=_("Open Resource"),
                label=_("Open Resource"),
                icon=WBIcon.LINK.icon,
                key="open_external_resource",
            ),
        }

    def get_custom_list_instance_buttons(self) -> set[WidgetButton]:
        return self.get_custom_instance_buttons()


class NotificationDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return ListDisplay(
            fields=[
                Field(key="notification_type", label="Type"),
                Field(key="title", label="Title"),
                Field(key="body", label="body"),
                Field(key="sent", label="Sent"),
                Field(key="read", label="Read"),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["notification_type", ".", "."],
                ["title", "sent", "read"],
                ["body", "body", "body"],
            ]
        )



# Generated by Django 4.2.6 on 2023-11-02 08:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0005_entry_salutation"),
    ]

    operations = [
        migrations.AddField(
            model_name="employeremployeerelationship",
            name="position_name",
            field=models.CharField(
                blank=True,
                help_text="If the selected position isn't an exact match, feel free to use this field to provide more details about the position. You can edit this field once the position is confirmed.",
                max_length=128,
                null=True,
                verbose_name="Alternative Position Name",
            ),
        ),
    ]



class Middleware:
    MIDDLEWARE = [
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "corsheaders.middleware.CorsMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
        # "maintenance_mode.middleware.MaintenanceModeMiddleware", # we leave it commented for now to try out the performance gain to not evaluate maintenance mode on every request lifecycle
    ]


class DevMiddleware:
    @property
    def MIDDLEWARE(self):
        middleware = []
        middleware_list = Middleware.MIDDLEWARE.copy()
        if self.DEBUG:
            middleware.append("debug_toolbar.middleware.DebugToolbarMiddleware")
        middleware_list.insert(
            middleware_list.index("django.contrib.sessions.middleware.SessionMiddleware") + 1,
            "django.middleware.locale.LocaleMiddleware",
        )  # The django "LocaleMiddleware" needs to come after "SessionMiddleware" and before "CommonMiddleware".
        middleware.extend(middleware_list)
        return middleware



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem

NotificationTypeMenuItem = MenuItem(
    label=_("Notification Types"),
    endpoint="wbcore:notifications:notification_type_setting-list",
    permission=ItemPermission(permissions=["notifications.view_notificationtype"]),
)

NotificationMenuItem = MenuItem(
    label=_("Notifications"),
    endpoint="wbcore:notifications:notification-list",
    permission=ItemPermission(permissions=["notifications.view_notification"]),
)



# Generated by Django 4.1.9 on 2023-05-30 07:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0003_notificationusertoken_updated"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="body",
            field=models.TextField(blank=True, null=True),
        ),
    ]



import json

import plotly.graph_objects as go
from django.core.cache import cache
from django.db.models import QuerySet
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet, ViewSetMixin
from wbcore.contrib.io.viewset_mixins import ImportExportDRFMixin
from wbcore.enums import WidgetType
from wbcore.filters import DjangoFilterBackend
from wbcore.fsm.mixins import FSMViewSetMixin
from wbcore.metadata.mixins import WBCoreMetadataConfigViewMixin
from wbcore.pagination import CursorPagination, LimitOffsetPagination
from wbcore.utils.renderers import BrowsableAPIRendererWithoutForms

from ..cache.mixins import CacheMixin
from .encoders import PandasDataCustomRenderer
from .generics import GenericAPIView
from .mixins import (
    ActionMixin,
    CreateModelMixin,
    DestroyModelMixin,
    DestroyMultipleModelMixin,
    DocumentationMixin,
    FilterMixin,
    ListModelMixin,
    MessageMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericViewSet(ViewSetMixin, ActionMixin, GenericAPIView):
    pass


class ViewSet(WBCoreMetadataConfigViewMixin, DocumentationMixin, ActionMixin, ViewSet):
    def get_serializer(self):
        if hasattr(self, "serializer_class"):
            return self.serializer_class()
        return None

    def get_serializer_class(self):
        return getattr(self, "serializer_class", None)


class ReadOnlyModelViewSet(
    ImportExportDRFMixin,
    DocumentationMixin,
    WBCoreMetadataConfigViewMixin,
    RetrieveModelMixin,
    ListModelMixin,
    FilterMixin,
    GenericViewSet,
):
    pagination_class = LimitOffsetPagination
    READ_ONLY = True


class ModelViewSet(
    ImportExportDRFMixin,
    DocumentationMixin,
    FSMViewSetMixin,
    WBCoreMetadataConfigViewMixin,
    CreateModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    DestroyMultipleModelMixin,
    ListModelMixin,
    FilterMixin,
    GenericViewSet,
):
    pagination_class = LimitOffsetPagination


class ReadOnlyInfiniteModelViewSet(ReadOnlyModelViewSet):
    pagination_class = None


class RepresentationViewSet(ReadOnlyModelViewSet):
    WIDGET_TYPE = WidgetType.SELECT.value
    NO_CACHE = True
    pagination_class = CursorPagination


class InfiniteDataModelView(ModelViewSet):
    pagination_class = None


class ChartViewSet(MessageMixin, ImportExportDRFMixin, FilterMixin, CacheMixin, ViewSet):
    IMPORT_ALLOWED: bool = False  # Disabled by default for chart
    EXPORT_ALLOWED: bool = False  # Disabled by default for chart

    filter_backends = [DjangoFilterBackend]
    WIDGET_TYPE = WidgetType.CHART.value

    renderer_classes = [PandasDataCustomRenderer, BrowsableAPIRendererWithoutForms]

    CACHE_EMPTY_VALUE = dict()
    CACHE_MAIN_KEY = "figure_dict"

    def _get_dataframe(self) -> dict:
        figure_dict = dict()
        if self.cache_enabled:
            figure_dict = self._get_cached_res()
        if not figure_dict:
            self.extra_cache_kwargs = {}
            queryset = self.filter_queryset(self.get_queryset())
            figure = go.Figure()
            if queryset.exists():
                figure = self.get_plotly(queryset)
            figure_json = figure.to_json()  # we serialize to use the default PlotlyEncoder
            figure_dict = json.loads(
                figure_json
            )  # we reserialize to be able to hijack the figure config. This adds an extra steps of serialization/deserialization but the overhead is negligable.
            if self.cache_enabled:
                cached_res = self.serialize_cache_results(figure_dict)
                cache.set(self._get_cache_key(), cached_res, timeout=self._get_cache_timeout())
        return figure_dict

    def list(self, request: Request, *args, **kwargs):
        self.request = request
        figure_dict = self._get_dataframe()
        data = self.parse_figure_dict(figure_dict)
        data["messages"] = list(self._get_messages(request))
        return Response(data)

    def get_queryset(self):
        return self.queryset

    def parse_figure_dict(self, figure_dict: dict[str, any]) -> dict[str, any]:
        figure_dict["config"] = {"responsive": True, "displaylogo": False}
        figure_dict["useResizeHandler"] = True
        figure_dict["style"] = {"width": "100%", "height": "100%", "minHeight": "300px"}
        return figure_dict

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        for backend in list(self.filter_backends):
            queryset = backend().filter_queryset(self.request, queryset, self)
        return queryset



import pytest
from django.core.exceptions import ValidationError
from wbcore.models.fields import validate_year


class TestYearField:
    @pytest.mark.parametrize(
        ("input", "valid"),
        [
            (3, False),
            (24, False),
            (165, False),
            (41877, False),
            (2065, True),
            (1969, True),
            (1283, True),
            (None, True),
        ],
    )
    def test_year_validator(self, input, valid):
        if valid:
            validate_year(input)
        else:
            with pytest.raises(ValidationError):
                validate_year(input)
